# Copyright    2023  Xiaomi Corp.        (authors: Wei Kang)
#              2023  Binbin Zhang (binbzha@qq.com)
#              2023  Kaixun Huang
#              2023  Chengdong Liang (liangchengdong@mail.nwpu.edu.cn)
# See ../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils.tokenize_utils import tokenize_by_bpe_model
from typing import Dict, List, Tuple
from collections import deque


def tokenize(context_list_path, symbol_table, bpe_model=None):
    """ Read biasing list from the biasing list address, tokenize and convert it
        into token id
    """
    if bpe_model is not None:
        import sentencepiece as spm
        sp = spm.SentencePieceProcessor()
        sp.load(bpe_model)
    else:
        sp = None

    with open(context_list_path, "r") as fin:
        context_txts = fin.readlines()

    context_list = []
    for context_txt in context_txts:
        context_txt = context_txt.strip()

        labels = []
        tokens = []
        if bpe_model is not None:
            tokens = tokenize_by_bpe_model(sp, context_txt)
        else:
            for ch in context_txt:
                if ch == ' ':
                    ch = "▁"
                tokens.append(ch)
        for ch in tokens:
            if ch in symbol_table:
                labels.append(symbol_table[ch])
            elif '<unk>' in symbol_table:
                labels.append(symbol_table['<unk>'])
        context_list.append(labels)
    return context_list


class ContextState:
    """The state in ContextGraph"""

    def __init__(
        self,
        id: int,
        token: int,
        token_score: float,
        node_score: float,
        output_score: float,
        is_end: bool,
    ):
        """Create a ContextState.

        Args:
          id:
            The node id, only for visualization now. A node is in [0, graph.num_nodes).
            The id of the root node is always 0.
          token:
            The token id.
          token_score:
            The bonus for each token during decoding, which will hopefully
            boost the token up to survive beam search.
          node_score:
            The accumulated bonus from root of graph to current node, it will be
            used to calculate the score for fail arc.
          output_score:
            The total scores of matched phrases, sum of the node_score of all
            the output node for current node.
          is_end:
            True if current token is the end of a context.
        """
        self.id = id
        self.token = token
        self.token_score = token_score
        self.node_score = node_score
        self.output_score = output_score
        self.is_end = is_end
        self.next = {}
        self.fail = None
        self.output = None


class ContextGraph:
    """The ContextGraph is modified from Aho-Corasick which is mainly
    a Trie with a fail arc for each node.
    See https://en.wikipedia.org/wiki/Aho%E2%80%93Corasick_algorithm for more details
    of Aho-Corasick algorithm.

    A ContextGraph contains some words / phrases that we expect to boost their
    scores during decoding. If the substring of a decoded sequence matches the word / phrase  # noqa
    in the ContextGraph, we will give the decoded sequence a bonus to make it survive
    beam search.
    """

    def __init__(self,
                 context_list_path: str,
                 symbol_table: Dict[str, int],
                 bpe_model: str = None,
                 context_score: float = 6.0):
        """Initialize a ContextGraph with the given ``context_score``.

        A root node will be created (**NOTE:** the token of root is hardcoded to -1).

        Args:
          context_score:
            The bonus score for each token(note: NOT for each word/phrase, it means longer  # noqa
            word/phrase will have larger bonus score, they have to be matched though).
        """
        self.context_score = context_score
        self.context_list = tokenize(context_list_path, symbol_table,
                                     bpe_model)
        self.num_nodes = 0
        self.root = ContextState(
            id=self.num_nodes,
            token=-1,
            token_score=0,
            node_score=0,
            output_score=0,
            is_end=False,
        )
        self.root.fail = self.root
        self.build_graph(self.context_list)

    def build_graph(self, token_ids: List[List[int]]):
        """Build the ContextGraph from a list of token list.
        It first build a trie from the given token lists, then fill the fail arc
        for each trie node.

        See https://en.wikipedia.org/wiki/Trie for how to build a trie.

        Args:
          token_ids:
            The given token lists to build the ContextGraph, it is a list of token list,
            each token list contains the token ids for a word/phrase. The token id
            could be an id of a char (modeling with single Chinese char) or an id
            of a BPE (modeling with BPEs).
        """
        for tokens in token_ids:
            node = self.root
            for i, token in enumerate(tokens):
                if token not in node.next:
                    self.num_nodes += 1
                    is_end = i == len(tokens) - 1
                    node_score = node.node_score + self.context_score
                    node.next[token] = ContextState(
                        id=self.num_nodes,
                        token=token,
                        token_score=self.context_score,
                        node_score=node_score,
                        output_score=node_score if is_end else 0,
                        is_end=is_end,
                    )
                node = node.next[token]
        self._fill_fail_output()  # AC

    def _fill_fail_output(self):
        """This function fills the fail arc for each trie node, it can be computed
        in linear time by performing a breadth-first search starting from the root.
        See https://en.wikipedia.org/wiki/Aho%E2%80%93Corasick_algorithm for the
        details of the algorithm.
        """
        queue = deque()
        for token, node in self.root.next.items():
            node.fail = self.root
            queue.append(node)
        while queue:
            current_node = queue.popleft()
            for token, node in current_node.next.items():
                fail = current_node.fail
                if token in fail.next:
                    fail = fail.next[token]
                else:
                    fail = fail.fail
                    while token not in fail.next:
                        fail = fail.fail
                        if fail.token == -1:  # root
                            break
                    if token in fail.next:
                        fail = fail.next[token]
                node.fail = fail
                # fill the output arc
                output = node.fail
                while not output.is_end:
                    output = output.fail
                    if output.token == -1:  # root
                        output = None
                        break
                node.output = output
                node.output_score += 0 if output is None else output.output_score
                queue.append(node)

    def forward_one_step(self, state: ContextState,
                         token: int) -> Tuple[float, ContextState]:
        """Search the graph with given state and token.

        Args:
          state:
            The given token containing trie node to start.
          token:
            The given token.

        Returns:
          Return a tuple of score and next state.
        """
        node = None
        score = 0
        # token matched
        if token in state.next:
            node = state.next[token]
            score = node.token_score
        else:
            # token not matched
            # We will trace along the fail arc until it matches the token or reaching
            # root of the graph.
            node = state.fail
            while token not in node.next:
                node = node.fail
                if node.token == -1:  # root
                    break

            if token in node.next:
                node = node.next[token]

            # The score of the fail path
            score = node.node_score - state.node_score
        assert node is not None
        return (score + node.output_score, node)

    def finalize(self, state: ContextState) -> Tuple[float, ContextState]:
        """When reaching the end of the decoded sequence, we need to finalize
        the matching, the purpose is to subtract the added bonus score for the
        state that is not the end of a word/phrase.

        Args:
          state:
            The given state(trie node).

        Returns:
          Return a tuple of score and next state. If state is the end of a word/phrase
          the score is zero, otherwise the score is the score of a implicit fail arc
          to root. The next state is always root.
        """
        # The score of the fail arc
        score = -state.node_score
        return (score, self.root)



import os.path
import random


import numpy as np
import torch
from torch import nn
import time


def set_parameter_requires_grad(model, feature_extracting):
    """是否提取预训练模型的参数,如果提取,则该参数不再更新"""
    if feature_extracting:
        for param in model.parameters():
            param.requires_grad = False


def create_instance_by_name(module_name, class_name, **kwargs):
    module_meta = __import__(module_name, globals(), locals(), [class_name])
    class_meta = getattr(module_meta, class_name)
    obj = class_meta(**kwargs)
    return obj


def put_data_to_device(*args):
    device = args[-1]
    res = list()
    for i in range(len(args) - 1):
        X = args[i]
        res.append(X.to(device))
    return tuple(res)


def print_attributes_of_object(obj):
    """打印指定对象的属性名-属性值"""
    attribute_names = dir(obj)
    # 打印属性名和对应的属性值
    for name in attribute_names:
        if not name.startswith('__'):  # 排除特殊方法和属性
            value = getattr(obj, name)
            print(f"Attribute: {name}, Value: {value}")


def get_attributes_of_object_to_dict(obj):
    """打印指定对象的属性名-属性值"""
    dict = {}
    attribute_names = dir(obj)
    # 打印属性名和对应的属性值
    for name in attribute_names:
        if not name.startswith('__'):  # 排除特殊方法和属性
            value = getattr(obj, name)
            dict[name] = value
    return dict


def set_random_seed(seed, cuda=True):
    """
    设置训练的随机种子
    """
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)

    if cuda:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def grad_clipping(net, theta):
    """Clip the gradient."""
    if isinstance(net, nn.Module):
        params = [p for p in net.parameters() if p.requires_grad]
    else:
        params = net.params
    norm = torch.sqrt(sum(torch.sum((p.grad ** 2)) for p in params))
    if norm > theta:
        for param in params:
            param.grad[:] *= theta / norm


def get_model_param_num(model):
    """
    存在打印行为
    得到：
    总参数数量，在训练的参数数量"""
    total_num = sum(p.numel() for p in model.parameters())
    trainable_num = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print(f'total_num: {total_num / 1024 / 1024}M, trainable_num: {trainable_num / 1024 / 1024}M')
    return total_num, trainable_num


def get_output_dim_for_conv(input_dim, kernel_size, stride=1, padding=0, dilation=1):
    """
    得到卷积后的输出维度, 不区分几维度的卷积.
    正确性已得到验证
    """
    return (input_dim + 2 * padding - (dilation * (kernel_size - 1) + 1)) // stride + 1


def get_output_dim_for_multi_conv(input_dim: int, kernel_size: list, stride: list, padding: list, dilation: list):
    assert len(kernel_size) == len(stride) == len(padding) == len(dilation)
    res = input_dim
    for kernel_size, stride, padding, dilation in zip(kernel_size, stride, padding, dilation):
        res = get_output_dim_for_conv(res, kernel_size, stride, padding, dilation)
    return res


def get_receptive_field_size_for_continuous_conv(kernel_sizes: list, strides: list):
    """
    得到连续卷积后的感受野, 不区分几维度的卷积.
    第一层的感受野即为卷积核大小, 后面层的感受野受到前面层的感受野
    正确性得到验证
    .   .   .   .   .   . 第一层
    . . . . . . . . . . . . . . . . .输入
    """
    assert len(kernel_sizes) == len(strides)
    assert len(kernel_sizes) > 0 and len(strides) > 0
    res = kernel_sizes[0]
    for kernel_size, stride in zip(kernel_sizes[1:], strides[1:]):
        res = (kernel_size - 1) * stride + res
    return res



class FormalCheckpoint:
    model_name: str
    model_state_dict: dict
    optimizer_state_dict: dict
    scheduler_state_dict: dict
    epoch: int
    step: int

    @property
    def data_dict(self):
        return {
            "model_state_dict": self.model_state_dict,
            "optimizer_state_dict": self.optimizer_state_dict,
            "scheduler_state_dict": self.scheduler_state_dict,
            "epoch": self.epoch,
            "step": self.step,
        }


def save_checkpoint(model, optimizer, scheduler, epoch, step, sava_dir: str):
    checkpoint = FormalCheckpoint(
        model_name=model.__class__.__name__,
        model_state_dict=model.state_dict(),
        optimizer_state_dict=optimizer.state_dict(),
        scheduler_state_dict=scheduler.state_dict(),
        epoch=epoch,
        step=step,
    ) if scheduler is not None else FormalCheckpoint(
        model_name=model.__class__.__name__,
        model_state_dict=model.state_dict(),
        optimizer_state_dict=optimizer.state_dict(),
        scheduler_state_dict={},
        epoch=epoch,
        step=step,
    )
    now_time = time.strftime("%Y_%m_%d_%H_%M_%S", time.localtime())
    filename = f'checkpoint_{checkpoint.model_name}_epoch_{epoch}_step_{step}_{now_time}.pth'
    save_path = os.path.join(sava_dir, filename)
    torch.save(checkpoint.data_dict, save_path)


def load_checkpoint(model, optimizer, scheduler, pth_file_path, local_rank=0, is_distributed=True):
    checkpoint = torch.load(pth_file_path, map_location=torch.device(
        'cuda:' + str(local_rank) if local_rank >= 0 else 'cpu'))
    epoch = checkpoint['epoch']
    step = checkpoint['step']
    params_with_prefix = checkpoint['model_state_dict']
    if is_distributed:
        params_ready = params_with_prefix
    else:
        params_ready = {k.replace("module.", ""): v for k, v in params_with_prefix.items()}

    model.load_state_dict(params_ready)
    optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
    scheduler.load_state_dict(checkpoint['scheduler_state_dict'])
    return epoch, step


def get_model_structure_chart_by_netron(model: torch.nn.Module, input_data: torch.Tensor):
    import netron
    onnx_path = "./onnx_model_name.onnx"
    torch.onnx.export(model, input_data, onnx_path)
    netron.start(onnx_path)


def make_freeze_all_params(model):
    for param in model.parameters():
        param.requires_grad = False



from torch import nn
from .component import MultiHeadAttention
from .component import PositionalEncoding
from .component import PositionWiseFeedForward
from .component import padding_mask


class EncoderLayer(nn.Module):
    def __init__(self, model_dim: int = 512, num_heads: int = 8, ffn_dim: int = 2048, dropout: float = 0.0):
        super(EncoderLayer, self).__init__()
        self.multi_head_attention = MultiHeadAttention(model_dim, num_heads, dropout)
        self.position_wise_feed_forward = PositionWiseFeedForward(model_dim, ffn_dim, dropout)

    def forward(self, input, attn_mask):
        context, attention = self.multi_head_attention(input, input, input, attn_mask)
        output = self.position_wise_feed_forward(context)
        return output, attention


class Encoder(nn.Module):
    """
    input->
    inputs: (batch_size, seq_len)
    input_len: (batch_size, 1)
    output->
    output: (batch_size, seq_len, model_dim)
    attentions: list(num_layers, tensor[batch_size, h, seq_len, seq_len])
    """

    def __init__(self, vocab_size: int, max_seq_lens: int,
                 num_layers: int = 6, model_dim: int = 512, num_heads: int = 8, ffn_dim: int = 2048,
                 dropout: float = 0.0, padding_idx: int = 0):
        super(Encoder, self).__init__()
        self.seq_embedding = nn.Embedding(vocab_size, model_dim, padding_idx=padding_idx)
        self.pos_embedding = PositionalEncoding(model_dim, max_seq_lens)
        self.encoder_layers = nn.ModuleList(
            [EncoderLayer(model_dim, num_heads, ffn_dim, dropout) for _ in range(num_layers)])

    def forward(self, inputs, input_len):
        """
        input->
        inputs: (batch_size, seq_len)
        input_len: (batch_size, 1)
        output->
        output: (batch_size, seq_len, model_dim)
        attentions: list(num_layers, tensor[batch_size, h, seq_len, seq_len])
        """
        output = self.seq_embedding(inputs)
        output += self.pos_embedding(input_len)
        self_attention_mask = padding_mask(inputs, inputs)
        attentions = []
        for layer in self.encoder_layers:
            output, attn = layer(output, self_attention_mask)
            attentions.append(attn)
        return output, attentions



import torch
from torch import nn
from gxl_ai_utils import store_model
from gxl_ai_utils.run.runner import RunnerGxl
from gxl_ai_utils.store_data.store_data import DataStore
from gxl_ai_utils.store_model import store_model_name

if __name__ == "__main__":
    model = store_model.store_model.load_model(store_model_name.MNIST)
    optim = torch.optim.SGD(model.parameters(), lr=0.1)
    loss_f = nn.CrossEntropyLoss()
    train_loader, valid_loader = DataStore('mnist').get_dataloader()
    runner = RunnerGxl(model, optim, loss_f, train_loader, valid_loader)
    runner.run(15)



import math
from typing import Any, List, Tuple, Union
import torch

from ..transformer.search import DecodeResult
from ..utils.mask import (make_non_pad_mask, mask_finished_preds,
                              mask_finished_scores)


def _isChinese(ch: str):
    if '\u4e00' <= ch <= '\u9fff' or '\u0030' <= ch <= '\u0039' or ch == '@':
        return True
    return False


def _isAllChinese(word: Union[List[Any], str]):
    word_lists = []
    for i in word:
        cur = i.replace(' ', '')
        cur = cur.replace('</s>', '')
        cur = cur.replace('<s>', '')
        cur = cur.replace('<unk>', '')
        cur = cur.replace('<OOV>', '')
        word_lists.append(cur)

    if len(word_lists) == 0:
        return False

    for ch in word_lists:
        if _isChinese(ch) is False:
            return False
    return True


def _isAllAlpha(word: Union[List[Any], str]):
    word_lists = []
    for i in word:
        cur = i.replace(' ', '')
        cur = cur.replace('</s>', '')
        cur = cur.replace('<s>', '')
        cur = cur.replace('<unk>', '')
        cur = cur.replace('<OOV>', '')
        word_lists.append(cur)

    if len(word_lists) == 0:
        return False

    for ch in word_lists:
        if ch.isalpha() is False and ch != "'":
            return False
        elif ch.isalpha() is True and _isChinese(ch) is True:
            return False

    return True


def paraformer_beautify_result(tokens: List[str]) -> str:
    middle_lists = []
    word_lists = []
    word_item = ''

    # wash words lists
    for token in tokens:
        if token in ['<sos>', '<eos>', '<blank>']:
            continue
        else:
            middle_lists.append(token)

    # all chinese characters
    if _isAllChinese(middle_lists):
        for i, ch in enumerate(middle_lists):
            word_lists.append(ch.replace(' ', ''))

    # all alpha characters
    elif _isAllAlpha(middle_lists):
        for i, ch in enumerate(middle_lists):
            word = ''
            if '@@' in ch:
                word = ch.replace('@@', '')
                word_item += word
            else:
                word_item += ch
                word_lists.append(word_item)
                word_lists.append(' ')
                word_item = ''

    # mix characters
    else:
        alpha_blank = False
        for i, ch in enumerate(middle_lists):
            word = ''
            if _isAllChinese(ch):
                if alpha_blank is True:
                    word_lists.pop()
                word_lists.append(ch)
                alpha_blank = False
            elif '@@' in ch:
                word = ch.replace('@@', '')
                word_item += word
                alpha_blank = False
            elif _isAllAlpha(ch):
                word_item += ch
                word_lists.append(word_item)
                word_lists.append(' ')
                word_item = ''
                alpha_blank = True
            else:
                word_lists.append(ch)

    return ''.join(word_lists).strip()


def paraformer_greedy_search(
        decoder_out: torch.Tensor,
        decoder_out_lens: torch.Tensor) -> List[DecodeResult]:
    batch_size = decoder_out.shape[0]
    maxlen = decoder_out.size(1)
    topk_prob, topk_index = decoder_out.topk(1, dim=2)
    topk_index = topk_index.view(batch_size, maxlen)  # (B, maxlen)
    topk_prob = topk_prob.view(batch_size, maxlen)
    results = []
    topk_index = topk_index.cpu().tolist()
    topk_prob = topk_prob.cpu().tolist()
    decoder_out_lens = decoder_out_lens.cpu().numpy()
    # TODO(Mddct) times
    for (i, hyp) in enumerate(topk_index):
        confidence = 0.0
        tokens_confidence = []
        lens = decoder_out_lens[i]
        for logp in topk_prob[i][:lens]:
            tokens_confidence.append(math.exp(logp))
            confidence += logp
        r = DecodeResult(hyp[:lens],
                         tokens_confidence=tokens_confidence,
                         confidence=math.exp(confidence / lens))
        results.append(r)
    return results


def paraformer_beam_search(decoder_out: torch.Tensor,
                           decoder_out_lens: torch.Tensor,
                           beam_size: int = 10,
                           eos: int = -1) -> List[DecodeResult]:
    mask = make_non_pad_mask(decoder_out_lens)
    indices, _ = _batch_beam_search(decoder_out,
                                    mask,
                                    beam_size=beam_size,
                                    eos=eos)

    best_hyps = indices[:, 0, :].cpu()
    decoder_out_lens = decoder_out_lens.cpu()
    results = []
    # TODO(Mddct): scores, times etc
    for (i, hyp) in enumerate(best_hyps.tolist()):
        r = DecodeResult(hyp[:decoder_out_lens.numpy()[i]])
        results.append(r)
    return results


def _batch_beam_search(
    logit: torch.Tensor,
    masks: torch.Tensor,
    beam_size: int = 10,
    eos: int = -1,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """ Perform batch beam search

        Args:
            logit: shape (batch_size, seq_length, vocab_size)
            masks: shape (batch_size, seq_length)
            beam_size: beam size

        Returns:
            indices: shape (batch_size, beam_size, seq_length)
            log_prob: shape (batch_size, beam_size)

        """

    batch_size, seq_length, vocab_size = logit.shape
    masks = ~masks
    # beam search
    with torch.no_grad():
        # b,t,v
        log_post = torch.nn.functional.log_softmax(logit, dim=-1)
        # b,k
        log_prob, indices = log_post[:, 0, :].topk(beam_size, sorted=True)
        end_flag = torch.eq(masks[:, 0], 1).view(-1, 1)
        # mask predictor and scores if end
        log_prob = mask_finished_scores(log_prob, end_flag)
        indices = mask_finished_preds(indices, end_flag, eos)
        # b,k,1
        indices = indices.unsqueeze(-1)

        for i in range(1, seq_length):
            # b,v
            scores = mask_finished_scores(log_post[:, i, :], end_flag)
            # b,v -> b,k,v
            topk_scores = scores.unsqueeze(1).repeat(1, beam_size, 1)
            # b,k,1 + b,k,v -> b,k,v
            top_k_logp = log_prob.unsqueeze(-1) + topk_scores

            # b,k,v -> b,k*v -> b,k
            log_prob, top_k_index = top_k_logp.view(batch_size,
                                                    -1).topk(beam_size,
                                                             sorted=True)

            index = mask_finished_preds(top_k_index, end_flag, eos)

            indices = torch.cat([indices, index.unsqueeze(-1)], dim=-1)

            end_flag = torch.eq(masks[:, i], 1).view(-1, 1)

        indices = torch.fmod(indices, vocab_size)

    return indices, log_prob



from typing import List, Tuple

import torch
import torch.nn as nn
from torch import Tensor

from gxl_ai_utils.config.gxl_config import GxlNode

#  tdnn encoder
"""
encoder_config:
    layer_num: int
    output_size: 
"""


class Tdnn(nn.Module):
    def __init__(self, input_size, output_size,
                 num_layers,
                 output_size_list,
                 kernel_size_list,
                 dilation_size_list,):
        super(Tdnn, self).__init__()
        self.tdnn = nn.Sequential()
        self.difference_lens = 0
        for i in range(num_layers):
            self.tdnn.append(
                nn.Conv1d(
                    in_channels=input_size,
                    out_channels=output_size_list[i],
                    kernel_size=kernel_size_list[i],
                    dilation=dilation_size_list[i]
                )
            )
            self.tdnn.append(nn.ReLU(
                inplace=False))  # inplace=False：这表示是否进行原地操作。当设置为False时，
            # 模块会创建一个新的输出张量，而不会修改输入张量。当设置为True时，模块会在原地修改输入张量，节省内存空间。
            self.tdnn.append(nn.BatchNorm1d(output_size_list[i], affine=False))  # affine 是否应用仿射变换
            input_size = output_size_list[i]
            self.difference_lens += dilation_size_list[i] * (kernel_size_list[i] - 1)
        self.output_linear = nn.Linear(output_size_list[-1], output_size)

    def forward(self, x: torch.Tensor, x_lens: torch.Tensor) -> tuple[Tensor, Tensor]:
        x = x.permute(0, 2, 1)  # (N, T, C) -> (N, C, T)
        x = self.tdnn(x)
        x = x.permute(0, 2, 1)  # (N, C, T) -> (N, T, C)
        logits = self.output_linear(x)
        x_lens = x_lens - self.difference_lens
        return logits, x_lens


def test_tdnn():
    input_tensor = torch.rand(3, 92, 80)
    input_lens = torch.tensor([92, 91, 83])
    config = GxlNode.get_config_from_yaml("./config.yaml")
    print(config)
    num_layer = config.encoder_conf.layer_num
    output_size_list = config.encoder_conf.output_size_list
    kernel_size_list = config.encoder_conf.kernel_size_list
    dilation_size_list = config.encoder_conf.dilation_size_list
    model = Tdnn(80, 128, num_layer, output_size_list, kernel_size_list, dilation_size_list)
    output, lens2 = model(input_tensor, input_lens)
    print(output.shape)
    print(lens2.shape)
    print(lens2)
    print(model)





import torch
import torch.nn as nn


class GPTMLP(nn.Module):
    def __init__(self, embed_dim=768):
        super(GPTMLP, self).__init__()

        # Point-wise feedforward network
        self.c_fc = nn.Conv1d(embed_dim, embed_dim * 4, kernel_size=1, bias=True)
        self.c_proj = nn.Conv1d(embed_dim * 4, embed_dim, kernel_size=1, bias=True)

        # Activation function (GELU)
        self.act = nn.GELU()

        # Dropout layer
        self.dropout = nn.Dropout(p=0.1, inplace=False)

    def forward(self, x):
        # Input x: [batch_size, seq_length, embed_dim]

        # Apply feedforward network and activation
        h = self.act(self.c_fc(x.transpose(1, 2)))  # [batch_size, 4 * embed_dim, seq_length]
        h2 = self.c_proj(h)  # [batch_size, embed_dim, seq_length]

        # Apply dropout
        h2 = self.dropout(h2)

        # Add residual connection
        output = x + h2
        return output



import copy
import os
from typing import Optional, Tuple, Union

import torch
import torch.nn.functional as F

try:
    from espnet2.asr.specaug.specaug import SpecAug
except ImportError:
    SpecAug = None
from ..utils.mask import make_pad_mask, add_optional_chunk_mask


class OpenAIWhisperEncoder(torch.nn.Module):
    """Transformer-based Speech Encoder from OpenAI's Whisper Model:

    URL: https://github.com/openai/whisper
    """

    def __init__(
            self,
            dropout_rate: float = 0.0,
            whisper_model: str = "small",
            use_specaug: bool = False,
            specaug_conf: Optional[dict] = None,
            do_pad_trim: bool = False,
            use_lora: bool = False,  # add to use lora
            freeze: bool = False
    ):
        try:
            from ... import gxl_whisper as whisper
            from ...gxl_whisper.audio import HOP_LENGTH, N_FFT, N_SAMPLES
            N_MELS = 80
        except Exception as e:
            print("Error: whisper is not properly installed.")
            print(
                "Please install whisper with: cd ${MAIN_ROOT}/tools &&",
                "./installers/install_whisper.sh",
            )
            raise e

        super().__init__()

        self.n_fft = N_FFT  # 快速傅里叶变换的长度
        self.win_length = N_FFT
        self.hop_length = HOP_LENGTH
        self.n_mels = N_MELS

        self.mel_filters = whisper.audio.mel_filters

        # note that originally Whisper doesn't use dropouts
        self.dropout = torch.nn.Dropout(dropout_rate)

        if not use_lora:
            assert whisper_model in whisper.available_models()
            _model = whisper.load_model(whisper_model, download_root=os.environ.get('WHISPER_CACHE','/home/work_nfs7/xlgeng/workspace/whisper'),
                                        device='cpu')
        else:
            _model = None
        # add for lora
        # if use_lora:
        #     from espnet2.asr.whisper_lora import Whisper_lora
        #     from whisper import ModelDimensions
        #     from whisper import _ALIGNMENT_HEADS
        #     import loralib as lora
        #
        #     device = "cuda" if torch.cuda.is_available() else "cpu"
        #     checkpoint = torch.load(f"/home/work_nfs7/xlgeng/workspace/whisper/{whisper_model}.pt", map_location=device)
        #     dims = ModelDimensions(**checkpoint["dims"])
        #     _model = Whisper_lora(dims)
        #     _model.load_state_dict(checkpoint["model_state_dict"], strict=False)
        #     _model.set_alignment_heads(_ALIGNMENT_HEADS[whisper_model])
        #     lora.mark_only_lora_as_trainable(_model)
        # add for lora

        self.encoders = copy.deepcopy(_model.encoder)
        self.encoders.train()
        del _model
        if use_specaug:  # 频谱增强
            self.specaug = SpecAug(**specaug_conf)
        else:
            self.specaug = None

        self.do_pad_trim = do_pad_trim
        self.pad_samples = N_SAMPLES

    def gxl_forward(self,
                    xs: torch.Tensor,
                    xs_lens: torch.Tensor,):
        """
        xs: (B, T, D)
        xs_lens: (B,)
        :param xs:
        :param xs_lens:
        :return:
        """
        T = xs.size(1)
        # mask为真实值为true
        masks = ~make_pad_mask(xs_lens, T).unsqueeze(1)  # (B, 1, T)
        # if self.global_cmvn is not None:
        #     xs = self.global_cmvn(xs)
        inputs = xs.permute(0, 2, 1)
        x = F.gelu(self.encoders.conv1(inputs))
        x = F.gelu(self.encoders.conv2(x))
        x = x.permute(0, 2, 1)
        mask_pad = masks  # (B, 1, T/subsample_rate)
        for layer in self.encoders.blocks:
            xs = layer(xs, mask=mask_pad)
        if self.normalize_before:
            xs = self.after_norm(xs)
        # Here we assume the mask is not changed in encoder layers, so just
        # return the masks before encoder layers, and the masks will be used
        # for cross attention with decoder later
        return xs, masks

    def output_size(self) -> int:
        return self.encoders.ln_post.normalized_shape[-1]

    def pad_or_trim(
            self,
            array: torch.Tensor,
            length: int,
            axis: int = -1,
    ) -> torch.Tensor:
        """Pad or trim the audio array to N_SAMPLES.

        Used in zero-shot inference cases.
        """
        if array.shape[axis] > length:
            array = array.index_select(
                dim=axis, index=torch.arange(length).to(array.device)
            )

        if array.shape[axis] < length:
            pad_widths = [(0, 0)] * array.ndim
            pad_widths[axis] = (0, length - array.shape[axis])
            array = F.pad(array, [pad for sizes in pad_widths[::-1] for pad in sizes])

        return array

    def log_mel_spectrogram(
            self,
            audio: torch.Tensor,
            ilens: torch.Tensor = None,
    ) -> torch.Tensor:
        """Use log-mel spectrogram computation native to Whisper training"""
        window = torch.hann_window(self.win_length).to(audio.device)
        stft = torch.stft(
            audio, self.n_fft, self.hop_length, window=window, return_complex=True
        )

        # whisper deletes the last frame by default (Shih-Lun)
        magnitudes = stft[..., :-1].abs() ** 2

        filters = self.mel_filters(audio.device, self.n_mels)
        mel_spec = filters @ magnitudes

        log_spec = torch.clamp(mel_spec, min=1e-10).log10()

        if ilens is not None:
            olens = ilens // self.hop_length
        else:
            olens = None

        log_spec = torch.maximum(
            log_spec,
            log_spec.view(audio.size(0), -1).max(dim=-1)[0][:, None, None] - 8.0,
        )
        log_spec = (log_spec + 4.0) / 4.0

        return log_spec, olens

    def whisper_encode(
            self,
            input: torch.Tensor,
            ilens: torch.Tensor = None,
    ) -> torch.Tensor:
        x = F.gelu(self.encoders.conv1(input))
        x = F.gelu(self.encoders.conv2(x))
        x = x.permute(0, 2, 1)

        n_frames = x.size(1)
        max_pos = self.encoders.positional_embedding.size(0)
        if n_frames <= max_pos:
            x = (x + self.encoders.positional_embedding[: x.size(1), :]).to(x.dtype)
        else:
            # due to positional encoding, audios >30 sec won't be accepted
            x = x[:, :max_pos, :] + self.encoders.positional_embedding

        x = self.dropout(x)

        for layer, block in enumerate(self.encoders.blocks):
            x = block(x)
            if layer < len(self.encoders.blocks) - 1:
                x = self.dropout(x)

        x = self.encoders.ln_post(x)

        if ilens is not None:
            olens = (
                    1
                    + (
                            ilens
                            - self.encoders.conv2.kernel_size[0]
                            + 2 * self.encoders.conv2.padding[0]
                    )
                    // self.encoders.conv2.stride[0]
            )
            olens = torch.clamp(olens, max=max_pos)
            T = x.size(2)
            masks = ~make_pad_mask(olens, T).unsqueeze(1)
        else:
            masks = None

        return x, masks

    def forward(
            self,
            xs_pad: torch.Tensor,
            ilens: torch.Tensor,
            decoding_chunk_size: int = 0,
            num_decoding_left_chunks: int = -1,
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        if self.do_pad_trim:
            xs_pad = self.pad_or_trim(xs_pad, self.pad_samples)

        feats, feats_lens = self.log_mel_spectrogram(xs_pad, ilens)
        raw_shape = xs_pad.shape
        mel_shape = feats.shape

        if self.specaug is not None and self.encoders.training:
            feats = feats.reshape(mel_shape[0], mel_shape[2], mel_shape[1])
            feats, feats_lens = self.specaug(feats, feats_lens)
            feats = feats.reshape(mel_shape[0], mel_shape[1], mel_shape[2])

        xs_pad, masks = self.whisper_encode(feats, feats_lens)

        return xs_pad, masks



# Copyright (c) 2020 Johns Hopkins University (Shinji Watanabe)
#               2020 Northwestern Polytechnical University (Pengcheng Guo)
#               2020 Mobvoi Inc (Binbin Zhang)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Swish() activation function for Conformer."""

import torch


class Swish(torch.nn.Module):
    """Construct an Swish object."""
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Return Swish activation function."""
        return x * torch.sigmoid(x)



# Copyright (c) 2020 Mobvoi Inc. (authors: Binbin Zhang, Di Wu)
#               2023 ASLP@NWPU (authors: He Wang, Fan Yu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified from ESPnet(https://github.com/espnet/espnet) and
# FunASR(https://github.com/alibaba-damo-academy/FunASR)

from typing import Dict, List, Optional, Tuple

import torch
from ..paraformer.cif import Cif

from ..paraformer.layers import SanmDecoder, SanmEncoder
from ..paraformer.layers import LFR
from ..paraformer.search import (paraformer_beam_search,
                                     paraformer_greedy_search)
from ..transformer.ctc import CTC
from ..transformer.decoder import TransformerDecoder
from ..transformer.encoder import BaseEncoder
from ..transformer.label_smoothing_loss import LabelSmoothingLoss
from ..transformer.search import DecodeResult
from ..utils.common import add_sos_eos
from ..utils.mask import make_non_pad_mask


class Paraformer(torch.nn.Module):
    """ Paraformer: Fast and Accurate Parallel Transformer for
        Non-autoregressive End-to-End Speech Recognition
        see https://arxiv.org/pdf/2206.08317.pdf

    """

    def __init__(self,
                 vocab_size: int,
                 encoder: BaseEncoder,
                 decoder: TransformerDecoder,
                 predictor: Cif,
                 ctc: Optional[CTC] = None,
                 ctc_weight: float = 0.5,
                 ignore_id: int = -1,
                 lsm_weight: float = 0,
                 length_normalized_loss: bool = False,
                 sampler: bool = True,
                 sampling_ratio: float = 0.75,
                 add_eos: bool = True,
                 **kwargs):
        assert isinstance(encoder,
                          SanmEncoder), isinstance(decoder, SanmDecoder)
        super().__init__()
        self.ctc_weight = ctc_weight
        self.ctc = ctc
        if ctc_weight == 0.0:
            del ctc
        self.encoder = encoder
        self.decoder = decoder
        self.predictor = predictor

        self.lfr = LFR()

        self.sos = 1
        self.eos = 2
        self.ignore_id = ignore_id

        self.criterion_att = LabelSmoothingLoss(
            size=vocab_size,
            padding_idx=ignore_id,
            smoothing=lsm_weight,
            normalize_length=length_normalized_loss)

        self.sampler = sampler
        self.sampling_ratio = sampling_ratio
        if sampler:
            self.embed = self.decoder.embed
        else:
            del self.decoder.embed
        # NOTE(Mddct): add eos in tail of labels for predictor
        # eg:
        #    gt:         你 好 we@@ net
        #    labels:     你 好 we@@ net eos
        self.add_eos = add_eos

    @torch.jit.ignore(drop=True)
    def forward(
        self,
        speech: torch.Tensor,
        speech_lengths: torch.Tensor,
        text: torch.Tensor,
        text_lengths: torch.Tensor,
    ) -> Dict[str, Optional[torch.Tensor]]:
        """Frontend + Encoder + Decoder + Calc loss

        Args:
            speech: (Batch, Length, ...)
            speech_lengths: (Batch, )
            text: (Batch, Length)
            text_lengths: (Batch,)
        """
        features, features_lens = self.lfr(speech, speech_lengths)
        features_lens = features_lens.to(speech_lengths.dtype)

        # 0 encoder
        encoder_out, encoder_out_mask = self.encoder(features, features_lens)

        # 1 predictor
        ys_pad, ys_pad_lens = text, text_lengths
        if self.add_eos:
            _, ys_pad = add_sos_eos(text, self.sos, self.eos, self.ignore_id)
            ys_pad_lens = text_lengths + 1
        acoustic_embd, _, _, _ = self.predictor(encoder_out, ys_pad,
                                                encoder_out_mask,
                                                self.ignore_id)

        # 2 decoder with sampler
        # TODO(Mddct): support mwer here
        acoustic_embd = self._sampler(
            encoder_out,
            encoder_out_mask,
            ys_pad,
            ys_pad_lens,
            acoustic_embd,
        )
        decoder_out, _, _ = self.decoder(encoder_out, encoder_out_mask,
                                         acoustic_embd, ys_pad_lens)

        # 3 loss
        # 3.1 ctc branhch
        loss_ctc: Optional[torch.Tensor] = None
        if self.ctc_weight != 0.0:
            loss_ctc = self._forward_ctc(encoder_out, encoder_out_mask, text,
                                         text_lengths)
        # TODO(Mddc): thu acc
        loss_decoder = self.criterion_att(decoder_out, ys_pad)
        loss = loss_decoder
        if loss_ctc is not None:
            loss = loss + self.ctc_weight * loss_ctc
        return {
            "loss": loss,
            "loss_ctc": loss_ctc,
            "loss_decoder": loss_decoder
        }

    @torch.jit.ignore(drop=True)
    def _forward_ctc(self, encoder_out: torch.Tensor,
                     encoder_mask: torch.Tensor, text: torch.Tensor,
                     text_lengths: torch.Tensor) -> torch.Tensor:
        encoder_out_lens = encoder_mask.squeeze(1).sum(1)
        loss_ctc = self.ctc(encoder_out, encoder_out_lens, text, text_lengths)
        return loss_ctc

    @torch.jit.ignore(drop=True)
    def _sampler(self, encoder_out, encoder_out_mask, ys_pad, ys_pad_lens,
                 pre_acoustic_embeds):
        device = encoder_out.device
        B, _ = ys_pad.size()

        tgt_mask = make_non_pad_mask(ys_pad_lens)
        ys_pad_embed = self.embed(ys_pad)  # [B, T, L]
        with torch.no_grad():
            decoder_out, _, _ = self.decoder(encoder_out, encoder_out_mask,
                                             pre_acoustic_embeds, ys_pad_lens)
            pred_tokens = decoder_out.argmax(-1)
            nonpad_positions = tgt_mask
            same_num = ((pred_tokens == ys_pad) * nonpad_positions).sum(1)
            input_mask = torch.ones_like(
                nonpad_positions,
                device=device,
                dtype=tgt_mask.dtype,
            )
            for li in range(B):
                target_num = (ys_pad_lens[li] -
                              same_num[li].sum()).float() * self.sampling_ratio
                target_num = target_num.long()
                if target_num > 0:
                    input_mask[li].scatter_(
                        dim=0,
                        index=torch.randperm(ys_pad_lens[li],
                                             device=device)[:target_num],
                        value=0,
                    )
            input_mask = torch.where(input_mask > 0, 1, 0)
            input_mask = input_mask * tgt_mask
            input_mask_expand = input_mask.unsqueeze(2)  # [B, T, 1]

        sematic_embeds = torch.where(input_mask_expand == 1,
                                     pre_acoustic_embeds, ys_pad_embed)
        # zero out the paddings
        return sematic_embeds * tgt_mask.unsqueeze(2)

    @torch.jit.export
    def forward_paraformer(
        self,
        speech: torch.Tensor,
        speech_lengths: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        features, features_lens = self.lfr(speech, speech_lengths)
        features_lens = features_lens.to(speech_lengths.dtype)
        # encoder
        encoder_out, encoder_out_mask = self.encoder(features, features_lens)

        # cif predictor
        acoustic_embed, token_num, _, _ = self.predictor(encoder_out,
                                                         mask=encoder_out_mask)
        token_num = token_num.floor().to(speech_lengths.dtype)

        # decoder
        decoder_out, _, _ = self.decoder(encoder_out, encoder_out_mask,
                                         acoustic_embed, token_num)
        decoder_out = decoder_out.log_softmax(dim=-1)
        return decoder_out, token_num

    def decode(self,
               methods: List[str],
               speech: torch.Tensor,
               speech_lengths: torch.Tensor,
               beam_size: int,
               decoding_chunk_size: int = -1,
               num_decoding_left_chunks: int = -1,
               ctc_weight: float = 0,
               simulate_streaming: bool = False,
               reverse_weight: float = 0) -> Dict[str, List[DecodeResult]]:
        decoder_out, decoder_out_lens = self.forward_paraformer(
            speech, speech_lengths)

        results = {}
        if 'paraformer_greedy_search' in methods:
            assert decoder_out is not None
            assert decoder_out_lens is not None
            paraformer_greedy_result = paraformer_greedy_search(
                decoder_out, decoder_out_lens)
            results['paraformer_greedy_search'] = paraformer_greedy_result
        if 'paraformer_beam_search' in methods:
            assert decoder_out is not None
            assert decoder_out_lens is not None
            paraformer_beam_result = paraformer_beam_search(
                decoder_out,
                decoder_out_lens,
                beam_size=beam_size,
                eos=self.eos)
            results['paraformer_beam_search'] = paraformer_beam_result

        return results



import random
from abc import ABC
from typing import Callable, Iterable

import torch
import torch.distributed as dist
from torch.utils.data import IterableDataset
from . import gxl_processor


class GxlDistributedSampler:
    def __init__(self, shuffle=True, partition=True):
        self.num_workers = None
        self.worker_id = None
        self.world_size = None
        self.rank = None
        self.epoch = -1
        self.update()
        self.shuffle = shuffle
        self.partition = partition

    def update(self):
        # assert dist.is_available()
        if not dist.is_available():
            self.worker_id = 0
            self.num_workers = 1
            self.worker_id = 0
            self.num_workers = 1
            return dict(rank=self.rank,
                        world_size=self.world_size,
                        worker_id=self.worker_id,
                        num_workers=self.num_workers)
        if dist.is_initialized():
            self.rank = dist.get_rank()
            self.world_size = dist.get_world_size()
        else:
            #  代表单卡系统, 多级系统会在训练代码中进行dict初始化
            self.rank = 0
            self.world_size = 1
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            self.worker_id = 0
            self.num_workers = 1
        else:
            self.worker_id = worker_info.id
            self.num_workers = worker_info.num_workers
        return dict(rank=self.rank,
                    world_size=self.world_size,
                    worker_id=self.worker_id,
                    num_workers=self.num_workers)

    def set_epoch(self, epoch):
        self.epoch = epoch

    def sample(self, data):
        """ Sample gxl_data according to rank/world_size/num_workers

            Args:
                data(List): input gxl_data list

            Returns:
                List: gxl_data list after sample
        """
        data = list(range(len(data)))
        # TODO(Binbin Zhang): fix this
        # We can not handle uneven gxl_data for CV on DDP, so we don't
        # sample gxl_data by rank, that means every GPU gets the same
        # and all the CV gxl_data
        if self.partition:
            if self.shuffle:
                random.Random(self.epoch).shuffle(data)
            data = data[self.rank::self.world_size]
        data = data[self.worker_id::self.num_workers]
        return data


class GxlDataList(IterableDataset, ABC):
    def __init__(self, lists: list, shuffle=True, partition=True):
        """
        对string_list加上一层sampler包装, 迭代的数据:
        {str: json_string, num_workers: 1, worker_id: 0, num_world:1, rank: 0}
        :param lists:  string_list
        :param shuffle:
        :param partition:
        """
        self.lists = lists
        self.sampler = GxlDistributedSampler(shuffle, partition)

    def set_epoch(self, epoch):
        self.sampler.set_epoch(epoch)

    def __iter__(self):
        sampler_info = self.sampler.update()
        indexes = self.sampler.sample(self.lists)
        for index in indexes:
            data = dict(src=self.lists[index])
            data.update(sampler_info)
            yield data


class GxlProcessor(IterableDataset, ABC):
    def __init__(self, source: Iterable, f: Callable, *args, **kw):
        """
        处理者: 使用处理函数进行处理
        :param source: 形如list的可迭代对象
        :param f: 处理函数
        :param args:
        :param kw:
        """
        assert callable(f)
        self.source = source
        self.f = f
        self.args = args
        self.kw = kw

    def set_epoch(self, epoch):
        self.source.set_epoch(epoch)

    def __iter__(self):
        """ Return an iterator over the source dataset processed by the
            given processor.
        """
        assert self.source is not None
        assert callable(self.f)
        return self.f(iter(self.source), *self.args, **self.kw)

    def apply(self, f):
        assert callable(f)
        return GxlProcessor(self, f, *self.args, **self.kw)


def GxlAsrDataset(data_type,
                  data_list_file,
                  symbol_table,
                  conf,
                  bpe_model=None,
                  non_lang_syms=None,
                  partition=True):
    """

    :param data_type:  in ['raw', 'shard']
    :param data_list_file:
    :param symbol_table:
    :param conf:
    :param bpe_model:
    :param non_lang_syms:
    :param partition:  whether to do gxl_data partition in terms of rank
    :return:
    """
    assert data_type in ['raw', 'shard']
    from ..utils.utils_file import load_list_file_clean
    lists = load_list_file_clean(data_list_file)
    shuffle = conf.get('shuffle', True)
    dataset = GxlDataList(lists, shuffle=shuffle, partition=partition)
    if data_type == 'shard':
        """"""
        # dataset = GxlProcessor(dataset, gxl_processor.url_opener)
        # dataset = Processor(dataset, processor.tar_file_and_group)
    else:
        dataset = GxlProcessor(dataset, gxl_processor.parse_raw)

    dataset = GxlProcessor(dataset, gxl_processor.tokenize, symbol_table, bpe_model,
                           non_lang_syms, conf.get('split_with_space', False))
    filter_conf = conf.get('filter_conf', {})
    dataset = GxlProcessor(dataset, gxl_processor.gxl_filter, **filter_conf)

    resample_conf = conf.get('resample_conf', {})
    dataset = GxlProcessor(dataset, gxl_processor.resample, **resample_conf)

    speed_perturb = conf.get('speed_perturb', False)
    if speed_perturb:
        dataset = GxlProcessor(dataset, gxl_processor.speed_perturb)

    feats_type = conf.get('feats_type', 'fbank')
    assert feats_type in ['fbank', 'mfcc']
    if feats_type == 'fbank':
        fbank_conf = conf.get('fbank_conf', {})
        dataset = GxlProcessor(dataset, gxl_processor.compute_fbank, **fbank_conf)
    elif feats_type == 'mfcc':
        mfcc_conf = conf.get('mfcc_conf', {})
        dataset = GxlProcessor(dataset, gxl_processor.compute_mfcc, **mfcc_conf)

    spec_aug = conf.get('spec_aug', True)
    spec_sub = conf.get('spec_sub', False)
    spec_trim = conf.get('spec_trim', False)
    if spec_aug:
        spec_aug_conf = conf.get('spec_aug_conf', {})
        dataset = GxlProcessor(dataset, gxl_processor.spec_augmentation, **spec_aug_conf)
    if spec_sub:
        spec_sub_conf = conf.get('spec_sub_conf', {})
        dataset = GxlProcessor(dataset, gxl_processor.spec_substitute, **spec_sub_conf)
    if spec_trim:
        spec_trim_conf = conf.get('spec_trim_conf', {})
        dataset = GxlProcessor(dataset, gxl_processor.spec_trim, **spec_trim_conf)

    if shuffle:
        shuffle_conf = conf.get('shuffle_conf', {})
        dataset = GxlProcessor(dataset, gxl_processor.shuffle, **shuffle_conf)

    sort = conf.get('sort', True)
    if sort:
        sort_conf = conf.get('sort_conf', {})
        dataset = GxlProcessor(dataset, gxl_processor.sort, **sort_conf)

    batch_conf = conf.get('batch_conf', {})
    dataset = GxlProcessor(dataset, gxl_processor.batch, **batch_conf)
    dataset = GxlProcessor(dataset, gxl_processor.padding)
    return dataset



import re
from os import PathLike
from typing import Dict, List, Union, Optional

from gxl_ai_utils.training.tokenizer.base_tokenizer import BaseTokenizer, T
from gxl_ai_utils.utils import utils_file

class CharTokenizer(BaseTokenizer):
    def __init__(self,
                 symbol_table: Union[str, PathLike, Dict],
                 non_lang_syms:Optional[Union[str, PathLike, List]]=None,
                 split_by_blank:bool=False,
                 connect_sym:str="",
                 unk_sym:str="<unk>"):
        """"""
        self.non_lang_pattern = None
        if non_lang_syms is not None:
            self.non_lang_pattern = re.compile(
                r"(\[[^\[\]]+\]|<[^<>]+>|{[^{}]+})"  # r，原始字符串， 不对\进行特殊转义
            )
            # 得到非语言序列
            if isinstance(non_lang_syms, List):
                self.non_lang_syms = non_lang_syms
            else:
                self.non_lang_syms = utils_file.load_list_file_clean(non_lang_syms)
        else:
            self.non_lang_syms = []
        # 得到symbol_table
        if isinstance(symbol_table, Dict):
            self._symbol_table = symbol_table
        else:
            self._symbol_table = utils_file.load_dict_from_scp_for_symbol_table(symbol_table)
        # 得到char_table
        self._chat_table = {v:k for k,v in self._symbol_table.items()}


        # 赋值类属性
        self.connect_sym = connect_sym
        self.unk_sym = unk_sym
        self.split_by_blank = split_by_blank



    def text2tokens(self, line: str) -> List[T]:
        # 为line洗个澡
        line = line.strip()
        # 对非语言符号进行切分
        if self.non_lang_pattern is not None:
            parts = self.non_lang_pattern.split(line.upper())
        else:
            parts = [line]
        # 对切分后的结果进行token提取
        tokens = []
        for part in parts:
            if len(part.strip())==0:
                continue
            if part in self.non_lang_syms:
                tokens.append(part)
            if self.split_by_blank:
                part = part.split(" ")
            for char in part:
                if char == " ":
                    char = "▁"
                tokens.append(char)
        return tokens

    def tokens2text(self, tokens: List[T]) -> str:
        return self.connect_sym.join(tokens)

    def ids2tokens(self, ids: List[int]) -> List[T]:
        return [self._chat_table[idx] for idx in ids]

    def tokens2ids(self, tokens: List[T]) -> List[int]:
        ids = []
        for ch in tokens:
            if ch in self._symbol_table:
                ids.append(self._symbol_table[ch])
            else:
                ids.append(self._symbol_table[self.unk_sym])
        return ids

    def vocab_size(self) -> int:
        return len(self._symbol_table)

    @property
    def symbol_table(self) -> Dict[T, int]:
        return self._symbol_table


def _test_re():
    non_lang_pattern = re.compile(
        r"(\[[^\[\]]+\]|<[^<>]+>|{[^{}]+})"  # r，原始字符串， 不对\进行特殊转义
    )
    test_str = "你好，我是耿雪龙， 我来自西北工业大学，喜喜，{{争吵}} ，不要呀啊啊啊[[[大哭]]] 呜呜，唉，别闹了<下雨了>快多于呀"
    res1 = non_lang_pattern.search(test_str)
    print(res1)
    print(res1.string)
    res2 = non_lang_pattern.findall(test_str)
    for item in res2:
        print(item)
    print(res2)
    res3 = non_lang_pattern.finditer(test_str)
    for item in res3:
        print(item)
    print(res3)
    res4 = non_lang_pattern.split(test_str)
    print(res4)
    for item in res4:
        print(item)

if __name__ =="__main__":
    """"""
    symbol_path = "./aishell.txt"
    tokenizer = CharTokenizer(
        symbol_table=symbol_path,
    )
    str_test = '我是耿雪龙，你好呀，无锡吸星大法，胜多负少发撒给大家十大发射点范德萨范文芳是滴，，发射点发生挂上为，说的都是，十大高手'
    tokens, ids = tokenizer.tokenize(str_test)
    print(tokens)
    print(ids)
    utils_file.print_dict(tokenizer.symbol_table)
    print(tokenizer.vocab_size())



import argparse
import distutils
import functools
import os
import tempfile
from collections import Counter
from typing import Union, Optional, Text, List

from jsonlines import jsonlines

from gxl_ai_utils import AiConstant
import sentencepiece as spm

logger = AiConstant.AI_LOGGER(AiConstant.LOG_PATH + "gxl_ai_utils.log")

IGNORE_ID = -1
# `sos` and `eos` using same token
SOS = "<eos>"
EOS = SOS
UNK = "<unk>"
BLANK = "<blank>"
MASKCTC = "<mask>"
SPACE = "<space>"


def load_dict(dict_path: Optional[Text], maskctc=False) -> Optional[List[Text]]:
    if dict_path is None:
        return None

    with open(dict_path, "r") as f:
        dictionary = f.readlines()
    # first token is `<blank>`
    # multi line: `<blank> 0\n`
    # one line: `<blank>`
    # space is relpace with <space>
    char_list = [entry[:-1].split(" ")[0] for entry in dictionary]
    if BLANK not in char_list:
        char_list.insert(0, BLANK)
    if EOS not in char_list:
        char_list.append(EOS)
    # for non-autoregressive maskctc model
    if maskctc and MASKCTC not in char_list:
        char_list.append(MASKCTC)
    return char_list


class TextFeaturizer():
    def __init__(self, unit_type, vocab, spm_model_prefix=None, maskctc=False):
        """Text featurizer, for processing or extracting features from text.txt.

        Currently, it supports char/word/sentence-piece level tokenizing and conversion into
        a list of token indices. Note that the token indexing order follows the
        given vocabulary file.

        Args:
            unit_type (str): unit type, e.g. char, word, spm, spm 指的是sentencepiece的方式
            vocab Option[str, list]: Filepath to load vocabulary for token indices conversion, or vocab list.
            spm_model_prefix (str, optional): spm model prefix. Defaults to None.
        """
        assert unit_type in ('char', 'spm', 'word')
        self.unit_type = unit_type
        self.unk = UNK
        self.maskctc = maskctc
        self.vocab_path_or_list = vocab

        if self.vocab_path_or_list:
            self.vocab_dict, self._id2token, self.vocab_list, self.unk_id, self.eos_id, self.blank_id = self._load_vocabulary_from_file(
                vocab, maskctc)
            self.vocab_size = len(self.vocab_list)
        else:
            logger.warning(
                "TextFeaturizer: not have vocab file or vocab list. Only Tokenizer can use, can not convert to token idx"
            )

        if unit_type == 'spm':
            spm_model = spm_model_prefix + '.model'
            self.sp = spm.SentencePieceProcessor()
            self.sp.Load(spm_model)

    def tokenize(self, text, replace_space=True):
        """tokenizer split text.txt into text.txt tokens"""
        if self.unit_type == 'char':
            tokens = self.char_tokenize(text, replace_space)
        elif self.unit_type == 'word':
            tokens = self.word_tokenize(text)
        else:  # spm
            tokens = self.spm_tokenize(text)
        return tokens

    def detokenize(self, tokens):
        """tokenizer convert text.txt tokens back to text.txt"""
        if self.unit_type == 'char':
            text = self.char_detokenize(tokens)
        elif self.unit_type == 'word':
            text = self.word_detokenize(tokens)
        else:  # spm
            text = self.spm_detokenize(tokens)
        return text

    def featurize(self, text):
        """Convert text.txt string to a list of token indices.

        Args:
            text (str): Text to process.

        Returns:
            List[int]: List of token indices.
        """
        assert self.vocab_path_or_list, "toidx need vocab path or vocab list"
        tokens = self.tokenize(text)
        ids = []
        for token in tokens:
            if token not in self.vocab_dict:
                logger.debug(f"Text Token: {token} -> {self.unk}")
                token = self.unk
            ids.append(self.vocab_dict[token])
        return ids

    def defeaturize(self, idxs):
        """Convert a list of token indices to text.txt string,
        ignore index after eos_id.

        Args:
            idxs (List[int]): List of token indices.

        Returns:
            str: Text.
        """
        assert self.vocab_path_or_list, "toidx need vocab path or vocab list"
        tokens = []
        for idx in idxs:
            if idx == self.eos_id:
                break
            tokens.append(self._id2token[idx])
        text = self.detokenize(tokens)
        return text

    def char_tokenize(self, text, replace_space=True):
        """Character tokenizer.

        Args:
            text (str): text.txt string.
            replace_space (bool): False only used by build_vocab.py.

        Returns:
            List[str]: tokens.
        """
        text = text.strip()
        if replace_space:
            tokens = [SPACE if item == " " else item for item in list(text)]
        else:
            tokens = list(text)
        return tokens

    def char_detokenize(self, tokens):
        """Character detokenizer.

        Args:
            tokens (List[str]): tokens.

        Returns:
           str: text.txt string.
        """
        tokens = [t.replace(SPACE, " ") for t in tokens]
        return "".join(tokens)

    def word_tokenize(self, text):
        """Word tokenizer, separate by <space>."""
        return text.strip().split()

    def word_detokenize(self, tokens):
        """Word detokenizer, separate by <space>."""
        return " ".join(tokens)

    def spm_tokenize(self, text):
        """spm tokenize.

        Args:
            text (str): text.txt string.

        Returns:
            List[str]: sentence pieces str code
        """
        stats = {"num_empty": 0, "num_filtered": 0}

        def valid(line):
            return True

        def encode(l):
            return self.sp.EncodeAsPieces(l)

        def encode_line(line):
            line = line.strip()
            if len(line) > 0:
                line = encode(line)
                if valid(line):
                    return line
                else:
                    stats["num_filtered"] += 1
            else:
                stats["num_empty"] += 1
            return None

        enc_line = encode_line(text)
        return enc_line

    def spm_detokenize(self, tokens, input_format='piece'):
        """spm detokenize.

        Args:
            ids (List[str]): tokens.

        Returns:
            str: text.txt
        """
        if input_format == "piece":

            def decode(l):
                return "".join(self.sp.DecodePieces(l))
        elif input_format == "id":

            def decode(l):
                return "".join(self.sp.DecodeIds(l))

        return decode(tokens)

    def _load_vocabulary_from_file(self, vocab: Union[str, list],
                                   maskctc: bool):
        """Load vocabulary from file."""
        if isinstance(vocab, list):
            vocab_list = vocab
        else:
            vocab_list = load_dict(vocab, maskctc)
        assert vocab_list is not None
        logger.debug(f"Vocab: {vocab_list}")

        id2token = dict(
            [(idx, token) for (idx, token) in enumerate(vocab_list)])
        token2id = dict(
            [(token, idx) for (idx, token) in enumerate(vocab_list)])

        blank_id = vocab_list.index(BLANK) if BLANK in vocab_list else -1
        maskctc_id = vocab_list.index(MASKCTC) if MASKCTC in vocab_list else -1
        unk_id = vocab_list.index(UNK) if UNK in vocab_list else -1
        eos_id = vocab_list.index(EOS) if EOS in vocab_list else -1
        sos_id = vocab_list.index(SOS) if SOS in vocab_list else -1
        space_id = vocab_list.index(SPACE) if SPACE in vocab_list else -1

        logger.debug(f"BLANK id: {blank_id}")
        logger.debug(f"UNK id: {unk_id}")
        logger.debug(f"EOS id: {eos_id}")
        logger.debug(f"SOS id: {sos_id}")
        logger.debug(f"SPACE id: {space_id}")
        logger.debug(f"MASKCTC id: {maskctc_id}")
        return token2id, id2token, vocab_list, unk_id, eos_id, blank_id


def count_manifest(counter, text_feature, manifest_path):
    manifest_jsons = []
    with jsonlines.open(manifest_path, 'r') as reader:
        for json_data in reader:
            manifest_jsons.append(json_data)

    for line_json in manifest_jsons:
        if isinstance(line_json['text.txt'], str):
            tokens = text_feature.tokenize(
                line_json['text.txt'], replace_space=False)

            counter.update(tokens)
        else:
            assert isinstance(line_json['text.txt'], list)
            for text in line_json['text.txt']:
                tokens = text_feature.tokenize(text, replace_space=False)
                counter.update(tokens)


def dump_text_manifest(fileobj, manifest_path, key='text.txt'):
    """
    得到jsonl文件(manifest)中的key元素,并按行写入文件
    """
    manifest_jsons = []
    with jsonlines.open(manifest_path, 'r') as reader:
        for json_data in reader:
            manifest_jsons.append(json_data)

    for line_json in manifest_jsons:
        if isinstance(line_json[key], str):
            fileobj.write(line_json[key] + "\n")
        else:
            assert isinstance(line_json[key], list)
            for line in line_json[key]:
                fileobj.write(line + "\n")


def build_vocab(manifest_paths="",
                vocab_path="examples/librispeech/gxl_data/vocab.txt",
                unit_type="char",
                count_threshold=0,
                text_keys='text.txt',
                spm_mode="unigram",
                spm_vocab_size=0,
                spm_model_prefix="",
                spm_character_coverage=0.9995):
    manifest_paths = [manifest_paths] if isinstance(manifest_paths,
                                                    str) else manifest_paths

    fout = open(vocab_path, 'w', encoding='utf-8')
    fout.write(BLANK + "\n")  # 0 will be used for "blank" in CTC
    fout.write(UNK + '\n')  # <unk> must be 1

    if unit_type == 'spm':
        # tools/spm_train --input=$wave_data/lang_char/input.txt
        # --vocab_size=${nbpe} --model_type=${bpemode}
        # --model_prefix=${bpemodel} --input_sentence_size=100000000
        import sentencepiece as spm

        fp = tempfile.NamedTemporaryFile(mode='w', delete=False)
        for manifest_path in manifest_paths:
            _text_keys = [text_keys] if type(
                text_keys) is not list else text_keys

            for text_key in _text_keys:
                dump_text_manifest(fp, manifest_path, key=text_key)
        fp.close()
        # train
        spm.SentencePieceTrainer.Train(
            input=fp.name,
            vocab_size=spm_vocab_size,
            model_type=spm_mode,
            model_prefix=spm_model_prefix,
            input_sentence_size=100000000,
            character_coverage=spm_character_coverage)
        os.unlink(fp.name)

    # encode
    text_feature = TextFeaturizer(unit_type, "", spm_model_prefix)
    counter = Counter()

    for manifest_path in manifest_paths:
        count_manifest(counter, text_feature, manifest_path)

    count_sorted = sorted(counter.items(), key=lambda x: x[1], reverse=True)
    tokens = []
    for token, count in count_sorted:
        if count < count_threshold:
            break
        # replace space by `<space>`
        token = SPACE if token == ' ' else token
        tokens.append(token)

    tokens = sorted(tokens)
    for token in tokens:
        fout.write(token + '\n')

    fout.write(SOS + "\n")  # <sos/eos>
    fout.close()

def add_arguments(argname, type, default, help, argparser, **kwargs):
    """Add argparse's argument.

    Usage:

    .. code-block:: python

        parser = argparse.ArgumentParser()
        add_argument("name", str, "Jonh", "User name.", parser)
        args = parser.parse_args()
    """
    type = distutils.util.strtobool if type == bool else type
    argparser.add_argument(
        "--" + argname,
        default=default,
        type=type,
        help=help + ' Default: %(default)s.',
        **kwargs)
def define_argparse():
    parser = argparse.ArgumentParser(description=__doc__)
    add_arg = functools.partial(add_arguments, argparser=parser)

    # yapf: disable
    add_arg('unit_type', str, "char", "Unit type, e.g. char, word, spm")
    add_arg('count_threshold', int, 0,
            "Truncation threshold for char/word counts.Default 0, no truncate.")
    add_arg('vocab_path', str,
            'examples/librispeech/gxl_data/vocab.txt',
            "Filepath to write the vocabulary.")
    add_arg('manifest_paths', str,
            None,
            "Filepaths of manifests for building vocabulary. "
            "You can provide multiple manifest files.",
            nargs='+',
            required=True)
    add_arg('text_keys', str,
            'text.txt',
            "keys of the text.txt in manifest for building vocabulary. "
            "You can provide multiple k.",
            nargs='+')
    # bpe
    add_arg('spm_vocab_size', int, 0, "Vocab size for spm.")
    add_arg('spm_mode', str, 'unigram', "spm model type, e.g. unigram, spm, char, word. only need when `unit_type` is spm")
    add_arg('spm_model_prefix', str, "", "spm_model_%(spm_mode)_%(count_threshold), spm model prefix, only need when `unit_type` is spm")
    add_arg('spm_character_coverage', float, 0.9995, "character coverage to determine the minimum symbols")
    # yapf: disable

    args = parser.parse_args()
    return args

def print_arguments(args, info=None):
    """Print argparse's arguments.

    Usage:

    .. code-block:: python

        parser = argparse.ArgumentParser()
        parser.add_argument("name", default="Jonh", type=str, help="User name.")
        args = parser.parse_args()
        print_arguments(args)

    :param args: Input argparse.Namespace for printing.
    :type args: argparse.Namespace
    """
    filename = ""
    if info:
        filename = info["__file__"]
    filename = os.path.basename(filename)
    print(f"----------- {filename} Configuration Arguments -----------")
    for arg, value in sorted(vars(args).items()):
        print("%s: %s" % (arg, value))
    print("-----------------------------------------------------------")

def main():
    args = define_argparse()
    print_arguments(args, globals())
    build_vocab(**vars(args))

if __name__ == '__main__':
    main()


# Copyright (c) 2019 Shigeki Karita
#               2020 Mobvoi Inc (Binbin Zhang)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch

'''
def subsequent_mask(
        size: int,
        device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Create mask for subsequent steps (size, size).

    This mask is used only in decoder which works in an auto-regressive mode.
    This means the current step could only do attention with its left steps.

    In encoder, fully attention is used when streaming is not necessary and
    the sequence is not long. In this  case, no attention mask is needed.

    When streaming is need, chunk-based attention is used in encoder. See
    subsequent_chunk_mask for the chunk-based attention mask.

    Args:
        size (int): size of mask
        str device (str): "cpu" or "cuda" or torch.Tensor.device
        dtype (torch.device): result dtype

    Returns:
        torch.Tensor: mask

    Examples:
        >>> subsequent_mask(3)
        [[1, 0, 0],
         [1, 1, 0],
         [1, 1, 1]]
    """
    ret = torch.ones(size, size, device=device, dtype=torch.bool)
    return torch.tril(ret)
'''

def subsequent_mask(
        size: int,
        device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Create mask for subsequent steps (size, size).

    This mask is used only in decoder which works in an auto-regressive mode.
    This means the current step could only do attention with its left steps.

    In encoder, fully attention is used when streaming is not necessary and
    the sequence is not long. In this  case, no attention mask is needed.

    When streaming is need, chunk-based attention is used in encoder. See
    subsequent_chunk_mask for the chunk-based attention mask.

    Args:
        size (int): size of mask
        str device (str): "cpu" or "cuda" or torch.Tensor.device
        dtype (torch.device): result dtype

    Returns:
        torch.Tensor: mask

    Examples:
        >>> subsequent_mask(3)
        [[1, 0, 0],
         [1, 1, 0],
         [1, 1, 1]]
    """
    arange = torch.arange(size, device=device)
    mask = arange.expand(size, size)
    arange = arange.unsqueeze(-1)
    mask = mask <= arange
    return mask


def subsequent_chunk_mask(
        size: int,
        chunk_size: int,
        num_left_chunks: int = -1,
        device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Create mask for subsequent steps (size, size) with chunk size,
       this is for streaming encoder

    Args:
        size (int): size of mask
        chunk_size (int): size of chunk
        num_left_chunks (int): number of left chunks
            <0: use full chunk
            >=0: use num_left_chunks
        device (torch.device): "cpu" or "cuda" or torch.Tensor.device

    Returns:
        torch.Tensor: mask

    Examples:
        >>> subsequent_chunk_mask(4, 2)
        [[1, 1, 0, 0],
         [1, 1, 0, 0],
         [1, 1, 1, 1],
         [1, 1, 1, 1]]
    """
    ret = torch.zeros(size, size, device=device, dtype=torch.bool)
    for i in range(size):
        if num_left_chunks < 0:
            start = 0
        else:
            start = max((i // chunk_size - num_left_chunks) * chunk_size, 0)
        ending = min((i // chunk_size + 1) * chunk_size, size)
        ret[i, start:ending] = True
    return ret


def add_optional_chunk_mask(xs: torch.Tensor, masks: torch.Tensor,
                            use_dynamic_chunk: bool,
                            use_dynamic_left_chunk: bool,
                            decoding_chunk_size: int, static_chunk_size: int,
                            num_decoding_left_chunks: int,
                            enable_full_context: bool = True):
    """ Apply optional mask for encoder.

    Args:
        xs (torch.Tensor): padded input, (B, L, D), L for max length
        mask (torch.Tensor): mask for xs, (B, 1, L)
        use_dynamic_chunk (bool): whether to use dynamic chunk or not
        use_dynamic_left_chunk (bool): whether to use dynamic left chunk for
            training.
        decoding_chunk_size (int): decoding chunk size for dynamic chunk, it's
            0: default for training, use random dynamic chunk.
            <0: for decoding, use full chunk.
            >0: for decoding, use fixed chunk size as set.
        static_chunk_size (int): chunk size for static chunk training/decoding
            if it's greater than 0, if use_dynamic_chunk is true,
            this parameter will be ignored
        num_decoding_left_chunks: number of left chunks, this is for decoding,
            the chunk size is decoding_chunk_size.
            >=0: use num_decoding_left_chunks
            <0: use all left chunks
        enable_full_context (bool):
            True: chunk size is either [1, 25] or full context(max_len)
            False: chunk size ~ U[1, 25]

    Returns:
        torch.Tensor: chunk mask of the input xs.
    """
    # Whether to use chunk mask or not
    if use_dynamic_chunk:
        max_len = xs.size(1)
        if decoding_chunk_size < 0:
            chunk_size = max_len
            num_left_chunks = -1
        elif decoding_chunk_size > 0:
            chunk_size = decoding_chunk_size
            num_left_chunks = num_decoding_left_chunks
        else:
            # chunk size is either [1, 25] or full context(max_len).
            # Since we use 4 times subsampling and allow up to 1s(100 frames)
            # delay, the maximum frame is 100 / 4 = 25.
            chunk_size = torch.randint(1, max_len, (1, )).item()
            num_left_chunks = -1
            if chunk_size > max_len // 2 and enable_full_context:
                chunk_size = max_len
            else:
                chunk_size = chunk_size % 25 + 1
                if use_dynamic_left_chunk:
                    max_left_chunks = (max_len - 1) // chunk_size
                    num_left_chunks = torch.randint(0, max_left_chunks,
                                                    (1, )).item()
        chunk_masks = subsequent_chunk_mask(xs.size(1), chunk_size,
                                            num_left_chunks,
                                            xs.device)  # (L, L)
        chunk_masks = chunk_masks.unsqueeze(0)  # (1, L, L)
        chunk_masks = masks & chunk_masks  # (B, L, L)
    elif static_chunk_size > 0:
        num_left_chunks = num_decoding_left_chunks
        chunk_masks = subsequent_chunk_mask(xs.size(1), static_chunk_size,
                                            num_left_chunks,
                                            xs.device)  # (L, L)
        chunk_masks = chunk_masks.unsqueeze(0)  # (1, L, L)
        chunk_masks = masks & chunk_masks  # (B, L, L)
    else:
        chunk_masks = masks
    return chunk_masks


def make_pad_mask(lengths: torch.Tensor, max_len: int = 0) -> torch.Tensor:
    """Make mask tensor containing indices of padded part.

    See description of make_non_pad_mask.

    Args:
        lengths (torch.Tensor): Batch of lengths (B,).
    Returns:
        torch.Tensor: Mask tensor containing indices of padded part.

    Examples:
        >>> lengths = [5, 3, 2]
        >>> make_pad_mask(lengths)
        masks = [[0, 0, 0, 0 ,0],
                 [0, 0, 0, 1, 1],
                 [0, 0, 1, 1, 1]]
    """
    batch_size = lengths.size(0)
    max_len = max_len if max_len > 0 else lengths.max().item()
    seq_range = torch.arange(0,
                             max_len,
                             dtype=torch.int64,
                             device=lengths.device)
    seq_range_expand = seq_range.unsqueeze(0).expand(batch_size, max_len)
    seq_length_expand = lengths.unsqueeze(-1)
    mask = seq_range_expand >= seq_length_expand
    return mask


def make_non_pad_mask(lengths: torch.Tensor) -> torch.Tensor:
    """Make mask tensor containing indices of non-padded part.

    The sequences in a batch may have different lengths. To enable
    batch computing, padding is need to make all sequence in same
    size. To avoid the padding part pass value to context dependent
    block such as attention or convolution , this padding part is
    masked.

    This pad_mask is used in both encoder and decoder.

    1 for non-padded part and 0 for padded part.

    Args:
        lengths (torch.Tensor): Batch of lengths (B,).
    Returns:
        torch.Tensor: mask tensor containing indices of padded part.

    Examples:
        >>> lengths = [5, 3, 2]
        >>> make_non_pad_mask(lengths)
        masks = [[1, 1, 1, 1 ,1],
                 [1, 1, 1, 0, 0],
                 [1, 1, 0, 0, 0]]
    """
    return ~make_pad_mask(lengths)


def mask_finished_scores(score: torch.Tensor,
                         flag: torch.Tensor) -> torch.Tensor:
    """
    If a sequence is finished, we only allow one alive branch. This function
    aims to give one branch a zero score and the rest -inf score.

    Args:
        score (torch.Tensor): A real value array with shape
            (batch_size * beam_size, beam_size).
        flag (torch.Tensor): A bool array with shape
            (batch_size * beam_size, 1).

    Returns:
        torch.Tensor: (batch_size * beam_size, beam_size).
    """
    beam_size = score.size(-1)
    zero_mask = torch.zeros_like(flag, dtype=torch.bool)
    if beam_size > 1:
        unfinished = torch.cat((zero_mask, flag.repeat([1, beam_size - 1])),
                               dim=1)
        finished = torch.cat((flag, zero_mask.repeat([1, beam_size - 1])),
                             dim=1)
    else:
        unfinished = zero_mask
        finished = flag
    score.masked_fill_(unfinished, -float('inf'))
    score.masked_fill_(finished, 0)
    return score


def mask_finished_preds(pred: torch.Tensor, flag: torch.Tensor,
                        eos: int) -> torch.Tensor:
    """
    If a sequence is finished, all of its branch should be <eos>

    Args:
        pred (torch.Tensor): A int array with shape
            (batch_size * beam_size, beam_size).
        flag (torch.Tensor): A bool array with shape
            (batch_size * beam_size, 1).

    Returns:
        torch.Tensor: (batch_size * beam_size).
    """
    beam_size = pred.size(-1)
    finished = flag.repeat([1, beam_size])
    return pred.masked_fill_(finished, eos)

def causal_or_lookahead_mask(
    mask: torch.Tensor,
    right_context: int,
    left_context: int,
    left_t_valid: int = 0,
) -> torch.Tensor:
    """Create mask (B, T, T) with history or future or both,
       this is for causal or noncausal streaming encoder

    Args:
        mask (torch.Tensor): size of mask shape (B, 1, T)
        right_context (int): future context size
        left_context (int): history context size
        left_t_valid (int): valid start offset

    Returns:
        torch.Tensor: mask shape (B, T, T)

    Examples:
        >>> seq_len  = torch.tensor([2,3,4])
        >>> seq_mask = make_non_pad_mask(seq_len)
        [[1, 1, 0, 0],
        [1, 1, 1, 0],
        [1, 1, 1, 1]]
        >>> causal_or_lookahead_mask(seq_mask.unsqueeze(1), 0, 2)
        [[[1, 0, 0, 0],
         [1, 1, 0, 0],
         [0, 0, 0, 0],
         [0, 0, 0, 0]],

        [[1, 0, 0, 0],
         [1, 1, 0, 0],
         [1, 1, 1, 0],
         [0, 0, 0, 0]],

        [[1, 0, 0, 0],
         [1, 1, 0, 0],
         [1, 1, 1, 0],
         [0, 1, 1, 1]]]
        >>> causal_or_lookahead_mask(seq_mask.unsqueeze(1), 1, 2)
        [[[1, 1, 0, 0],
         [1, 1, 0, 0],
         [0, 0, 0, 0],
         [0, 0, 0, 0]],

        [[1, 1, 0, 0],
         [1, 1, 1, 0],
         [1, 1, 1, 0],
         [0, 0, 0, 0]],

        [[1, 1, 0, 0],
         [1, 1, 1, 0],
         [1, 1, 1, 1],
         [0, 1, 1, 1]]]
    """
    _, _, T = mask.size()
    indices = torch.arange(T, device=mask.device)
    start = torch.where(indices > left_context, indices - left_context, 0)
    start = torch.where(indices < left_t_valid, indices, start).unsqueeze(1)

    end = indices + right_context + 1
    end = end.unsqueeze(1)
    indices_expand = indices.unsqueeze(0)
    gt = (indices_expand >= start)
    lt = (indices_expand < end)

    return (gt & lt) * mask.transpose(1, 2) * mask



from .transducer_search import greedy_search


def greedy_search_for_one_batch(input_x, x_lens, model):
    """

    :param input_x:
    :param input_lens:
    :param model:
    :return:
    """
    encoder_outputs, x_lens = model.encoder(input_x, x_lens)
    batch_size = encoder_outputs.size(0)
    res_list = []
    for i in range(batch_size):
        encoder_output_now = encoder_outputs[i:i+1]
        encoder_output_now = encoder_output_now[:, :x_lens[i], :]
        hyp = greedy_search(model, encoder_output_now)
        res_list.append(hyp)
    return res_list


def test_greedy_search():
    """"""






# Copyright (c) 2020 Mobvoi Inc. (authors: Binbin Zhang, Xiaoyu Chen, Di Wu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import argparse
import copy
import logging
import os

import torch
import yaml
from torch.utils.data import DataLoader

from ..gxl_dataset_wenet.gxl_dataset import GxlAsrDataset as Dataset
from ..gxl_model_wenet.utils.file_utils import read_symbol_table, read_non_lang_symbols
from ..gxl_model_wenet.utils.config import override_config
from ..gxl_model_wenet.init_model import init_model
from ..gxl_model_wenet.utils.context_graph import ContextGraph


def get_args():
    parser = argparse.ArgumentParser(description='recognize with your model')
    parser.add_argument('--config', required=True, help='config file')
    parser.add_argument('--test_data', required=True, help='test_gxl_ai_utils gxl_data file')
    parser.add_argument('--data_type',
                        default='raw',
                        choices=['raw', 'shard'],
                        help='train and cv gxl_data type')
    parser.add_argument('--gpu',
                        type=int,
                        default=-1,
                        help='gpu id for this rank, -1 for cpu')
    parser.add_argument('--checkpoint', required=True, help='checkpoint model')
    parser.add_argument('--dict', required=True, help='dict file')
    parser.add_argument(
        "--non_lang_syms",
        help="non-linguistic symbol file. One symbol per line.")
    parser.add_argument('--beam_size',
                        type=int,
                        default=10,
                        help='beam size for search')
    parser.add_argument('--penalty',
                        type=float,
                        default=0.0,
                        help='length penalty')
    parser.add_argument('--result_dir', required=True, help='asr result file')
    parser.add_argument('--batch_size',
                        type=int,
                        default=16,
                        help='asr result file')
    parser.add_argument('--modes',
                        nargs='+',
                        help="""decoding mode, support the following:
                             attention
                             ctc_greedy_search
                             ctc_prefix_beam_search
                             attention_rescoring
                             rnnt_greedy_search
                             rnnt_beam_search
                             rnnt_beam_attn_rescoring
                             ctc_beam_td_attn_rescoring
                             hlg_onebest
                             hlg_rescore
                             paraformer_greedy_search
                             paraformer_beam_search""")
    parser.add_argument('--search_ctc_weight',
                        type=float,
                        default=1.0,
                        help='ctc weight for nbest generation')
    parser.add_argument('--search_transducer_weight',
                        type=float,
                        default=0.0,
                        help='transducer weight for nbest generation')
    parser.add_argument('--ctc_weight',
                        type=float,
                        default=0.0,
                        help='ctc weight for rescoring weight in \
                                  attention rescoring decode mode \
                              ctc weight for rescoring weight in \
                                  transducer attention rescore decode mode')

    parser.add_argument('--transducer_weight',
                        type=float,
                        default=0.0,
                        help='transducer weight for rescoring weight in '
                        'transducer attention rescore mode')
    parser.add_argument('--attn_weight',
                        type=float,
                        default=0.0,
                        help='attention weight for rescoring weight in '
                        'transducer attention rescore mode')
    parser.add_argument('--decoding_chunk_size',
                        type=int,
                        default=-1,
                        help='''decoding chunk size,
                                <0: for decoding, use full chunk.
                                >0: for decoding, use fixed chunk size as set.
                                0: used for training, it's prohibited here''')
    parser.add_argument('--num_decoding_left_chunks',
                        type=int,
                        default=-1,
                        help='number of left chunks for decoding')
    parser.add_argument('--simulate_streaming',
                        action='store_true',
                        help='simulate streaming inference')
    parser.add_argument('--reverse_weight',
                        type=float,
                        default=0.0,
                        help='''right to left weight for attention rescoring
                                decode mode''')
    parser.add_argument('--bpe_model',
                        default=None,
                        type=str,
                        help='bpe model for english part')
    parser.add_argument('--override_config',
                        action='append',
                        default=[],
                        help="override yaml config")
    parser.add_argument('--connect_symbol',
                        default='',
                        type=str,
                        help='used to connect the output characters')

    parser.add_argument('--word',
                        default='',
                        type=str,
                        help='word file, only used for hlg decode')
    parser.add_argument('--hlg',
                        default='',
                        type=str,
                        help='hlg file, only used for hlg decode')
    parser.add_argument('--lm_scale',
                        type=float,
                        default=0.0,
                        help='lm scale for hlg attention rescore decode')
    parser.add_argument('--decoder_scale',
                        type=float,
                        default=0.0,
                        help='lm scale for hlg attention rescore decode')
    parser.add_argument('--r_decoder_scale',
                        type=float,
                        default=0.0,
                        help='lm scale for hlg attention rescore decode')

    parser.add_argument(
            '--context_bias_mode',
        type=str,
        default='',
        help='''Context bias mode, selectable from the following
                                option: decoding-graph、deep-biasing''')
    parser.add_argument('--context_list_path',
                        type=str,
                        default='',
                        help='Context list path')
    parser.add_argument('--context_graph_score',
                        type=float,
                        default=0.0,
                        help='''The higher the score, the greater the degree of
                                bias using decoding-graph for biasing''')
    args = parser.parse_args()
    print(args)
    return args


def main():
    args = get_args()
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s %(message)s')
    os.environ['CUDA_VISIBLE_DEVICES'] = str(args.gpu)

    with open(args.config, 'r') as fin:
        configs = yaml.load(fin, Loader=yaml.FullLoader)
    if len(args.override_config) > 0:
        configs = override_config(configs, args.override_config)

    symbol_table = read_symbol_table(args.dict)
    test_conf = copy.deepcopy(configs['dataset_conf'])

    test_conf['filter_conf']['max_length'] = 102400
    test_conf['filter_conf']['min_length'] = 0
    test_conf['filter_conf']['token_max_length'] = 102400
    test_conf['filter_conf']['token_min_length'] = 0
    test_conf['filter_conf']['max_output_input_ratio'] = 102400
    test_conf['filter_conf']['min_output_input_ratio'] = 0
    test_conf['speed_perturb'] = False
    test_conf['spec_aug'] = False
    test_conf['spec_sub'] = False
    test_conf['spec_trim'] = False
    test_conf['shuffle'] = False
    test_conf['sort'] = False
    if 'fbank_conf' in test_conf:
        test_conf['fbank_conf']['dither'] = 0.0
    elif 'mfcc_conf' in test_conf:
        test_conf['mfcc_conf']['dither'] = 0.0
    test_conf['batch_conf']['batch_type'] = "static"
    test_conf['batch_conf']['batch_size'] = args.batch_size
    non_lang_syms = read_non_lang_symbols(args.non_lang_syms)

    test_dataset = Dataset(args.data_type,
                           args.test_data,
                           symbol_table,
                           test_conf,
                           args.bpe_model,
                           non_lang_syms,
                           partition=False)

    test_data_loader = DataLoader(test_dataset, batch_size=None, num_workers=0)

    # Init asr model from configs
    model, configs = init_model(args, configs)

    # Load dict
    char_dict = {v: k for k, v in symbol_table.items()}

    use_cuda = args.gpu >= 0 and torch.cuda.is_available()
    device = torch.device('cuda' if use_cuda else 'cpu')
    model = model.to(device)
    model.eval()

    context_graph = None
    if 'decoding-graph' in args.context_bias_mode:
        context_graph = ContextGraph(args.context_list_path, symbol_table,
                                     args.bpe_model, args.context_graph_score)

    # TODO(Dinghao Zhou): Support RNN-T related decoding
    # TODO(Lv Xiang): Support k2 related decoding
    # TODO(Kaixun Huang): Support context graph
    files = {}
    for mode in args.modes:
        dir_name = os.path.join(args.result_dir, mode)
        os.makedirs(dir_name, exist_ok=True)
        file_name = os.path.join(dir_name, 'text.txt')
        files[mode] = open(file_name, 'w')
    max_format_len = max([len(mode) for mode in args.modes])
    with torch.no_grad():
        for batch_idx, batch in enumerate(test_data_loader):
            keys = batch["keys"]
            feats = batch["feats"].to(device)
            target = batch["target"].to(device)
            feats_lengths = batch["feats_lengths"].to(device)
            target_lengths = batch["target_lengths"].to(device)
            results = model.decode(
                args.modes,
                feats,
                feats_lengths,
                args.beam_size,
                decoding_chunk_size=args.decoding_chunk_size,
                num_decoding_left_chunks=args.num_decoding_left_chunks,
                ctc_weight=args.ctc_weight,
                simulate_streaming=args.simulate_streaming,
                reverse_weight=args.reverse_weight,
                context_graph=context_graph)
            for i, key in enumerate(keys):
                for mode, hyps in results.items():
                    content = [char_dict[w] for w in hyps[i].tokens]
                    line = '{} {}'.format(key,
                                          args.connect_symbol.join(content))
                    logging.info('{} {}'.format(mode.ljust(max_format_len),
                                                line))
                    files[mode].write(line + '\n')
    for mode, f in files.items():
        f.close()


if __name__ == '__main__':
    main()



import torch
from torch import nn
from .mlp_model import GPTMLP


class GxlAttention(nn.Module):
    def __init__(self, embed_dim=768, num_heads=12):
        super(GxlAttention, self).__init__()

        # Multi-Head Self Attention
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads

        # Query, Key, and Value projections
        self.c_attn = nn.Conv1d(embed_dim, embed_dim * 3, kernel_size=1, bias=False)
        self.c_proj = nn.Conv1d(embed_dim, embed_dim, kernel_size=1, bias=False)

        # Dropout layers
        self.attn_dropout = nn.Dropout(p=0.1, inplace=False)
        self.resid_dropout = nn.Dropout(p=0.1, inplace=False)

    def forward(self, x):
        # Input x: [batch_size, seq_length, embed_dim]

        batch_size, seq_length, embed_dim = x.size()

        # Project the input into query, key, and value
        qkv = self.c_attn(x.transpose(1, 2))  # [batch_size, 3 * embed_dim, seq_length]
        q, k, v = qkv.chunk(3, dim=1)  # Split into query, key, and value

        # Reshape for multi-head attention
        q = q.view(batch_size, self.num_heads, self.head_dim, seq_length).transpose(2,
                                                                                    3)  # [batch_size, num_heads, seq_length, head_dim]
        k = k.view(batch_size, self.num_heads, self.head_dim, seq_length).transpose(2,
                                                                                    3)  # [batch_size, num_heads, seq_length, head_dim]
        v = v.view(batch_size, self.num_heads, self.head_dim, seq_length).transpose(2,
                                                                                    3)  # [batch_size, num_heads, seq_length, head_dim]

        # Compute attention scores and apply softmax
        attn_scores = torch.matmul(q, k.transpose(-2, -1)) / (
                self.head_dim ** 0.5)  # [batch_size, num_heads, seq_length, seq_length]
        attn_probs = nn.functional.softmax(attn_scores, dim=-1)  # Apply softmax along the last dimension

        # Apply attention dropout
        attn_probs = self.attn_dropout(attn_probs)

        # Compute weighted sum of values
        attn_output = torch.matmul(attn_probs, v)  # [batch_size, num_heads, seq_length, head_dim]

        # Reshape and concatenate heads
        attn_output = attn_output.transpose(2, 3).contiguous().view(batch_size, seq_length,
                                                                    embed_dim)  # [batch_size, seq_length, embed_dim]

        # Project back to model's dimension
        attn_output = self.c_proj(attn_output.transpose(1, 2))  # [batch_size, embed_dim, seq_length]
        attn_output = self.resid_dropout(attn_output)

        # Add residual connection and LayerNorm
        output = x + attn_output
        return output


class GXLMLP(nn.Module):
    def __init__(self, embed_dim=768):
        super(GXLMLP, self).__init__()

        # Point-wise feedforward network
        self.c_fc = nn.Conv1d(embed_dim, embed_dim * 4, kernel_size=1, bias=True)
        self.c_proj = nn.Conv1d(embed_dim * 4, embed_dim, kernel_size=1, bias=True)
        self.act = nn.ReLU()
        # Dropout layer
        self.dropout = nn.Dropout(p=0.081024, inplace=False)

    def forward(self, x):
        # Input x: [batch_size, seq_length, embed_dim]

        # Apply feedforward network and activation
        h = self.act(self.c_fc(x.transpose(1, 2)))  # [batch_size, 4 * embed_dim, seq_length]
        h2 = self.c_proj(h)  # [batch_size, embed_dim, seq_length]

        # Apply dropout
        h2 = self.dropout(h2)

        # Add residual connection
        output = x + h2
        return output


class GPT2Block(nn.Module):
    def __init__(self):
        super(GPT2Block, self).__init__()
        self.ln_1 = nn.LayerNorm(768, eps=1e-05, elementwise_affine=True)
        self.attn = GxlAttention()
        self.ln_2 = nn.LayerNorm(768, eps=1e-05, elementwise_affine=True)
        self.mlp = GPTMLP()

    def forward(self, x):
        # Apply LayerNorm1
        x_normalized = self.ln_1(x)
        # Apply the self-attention mechanism (assuming self.attn is a nn.Module)
        attn_output = self.attn(x_normalized)  # Adjust this line based on input shape
        # Residual connection and LayerNorm2
        x_residual = x + attn_output
        x_normalized_residual = self.ln_2(x_residual)
        # Apply MLP
        mlp_output = self.mlp(x_normalized_residual)  # Adjust this line based on input shape
        # Residual connection and final output
        output = x_residual + mlp_output
        return output


class GXLTransformer(nn.Module):
    def __init__(self):
        super(GXLTransformer, self).__init__()
        self.wte = nn.Embedding(30000, 768)
        self.wpe = nn.Embedding(1024, 768)
        self.drop = nn.Dropout(p=0.1, inplace=False)
        self.h = nn.ModuleList([GPT2Block() for _ in range(12)])
        self.ln_f = nn.LayerNorm(768, eps=1e-05, elementwise_affine=True)

    def forward(self, input_ids):
        # input_ids shape: (batch_size, sequence_length)
        # Embedding
        wte_output = self.wte(input_ids)  # shape: (batch_size, sequence_length, 768)
        b, t, c = wte_output.size()
        # Positional encodings
        wpe_output = self.wpe(torch.arange(t, device=wte_output.device))[None, :, :].expand(b, t, c)
        # wpe_output shape: (batch_size, sequence_length, 768)

        # Combine embeddings and positional encodings
        hidden_states = wte_output + wpe_output
        # hidden_states shape: (batch_size, sequence_length, 768)

        # Apply dropout
        hidden_states = self.drop(hidden_states)
        # hidden_states shape: (batch_size, sequence_length, 768)

        # Loop through GPT2Blocks
        for block in self.h:
            hidden_states = block(hidden_states)
            # hidden_states shape: (batch_size, sequence_length, 768)

        # Apply LayerNorm
        output = self.ln_f(hidden_states)
        # output shape: (batch_size, sequence_length, 768)

        return output



import torch
from torch import nn
from gxl_ai_utils.utils import utils_file
import torchaudio
from .encoder import Tdnn
from .decoder import LSTMDecoder
from .joiner import Joiner
from .transducer_decode import greedy_search_for_one_batch

class Transducer(nn.Module):
    def __init__(self,
                 encoder,
                 decoder,
                 joiner):
        """"""
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.joiner = joiner

    def forward(self,
                x: torch.Tensor,
                x_lens: torch.Tensor,
                y: torch.Tensor,
                y_lens: torch.Tensor):
        """"""
        assert x.ndim == 3, x.ndim
        assert x_lens.ndim == 1, x_lens.ndim
        assert y.ndim == 2, y.shape[-1]
        assert y_lens.ndim == 1, y_lens.ndim

        blank_id = self.decoder.blank_id
        sos = torch.ones(x.shape[0], 1, dtype=torch.long) * blank_id
        sos_y = torch.cat([sos, y], dim=1)
        sos_y_padded = utils_file.do_padding_ids_by_lens(sos_y, x_lens, blank_id)
        decoder_out, _ = self.decoder(sos_y_padded)

        encoder_out, x_lens = self.encoder(x, x_lens)
        assert torch.all(x_lens >= 0), x_lens
        logits = self.joiner(encoder_out, decoder_out)
        # rnnt_loss need 0 padded targets
        y_padded_with_zero = utils_file.do_padding_ids_by_lens(y, y_lens, 0)
        print(logits.shape)
        print(y_padded_with_zero.shape)
        print(x_lens)
        print(y_lens)
        loss = torchaudio.functional.rnnt_loss(
            logits=logits,
            targets=y_padded_with_zero,
            logit_lengths=x_lens,
            target_lengths=y_lens,
            blank=blank_id,
            reduction='mean'
        )
        return loss

    def decode(self, x, x_lens):
        return greedy_search_for_one_batch(x, x_lens, self)


def test_transducer():
    """"""
    input_tensor = torch.randint(0, 1023, size=(3, 12))
    lengths = torch.randint(5, 12, size=(3,))
    print()
    print(input_tensor)
    print(lengths)
    output_y = utils_file.do_padding_ids_by_lens(input_tensor, lengths, 120)
    print()
    print(output_y)
    input_tensor = torch.randn(3, 5, 8)
    lengths = torch.randint(2, 6, size=(3,))
    print(lengths)
    output_y = utils_file.do_padding_embeds_by_lens(input_tensor, lengths, 10)
    print(output_y)

    x = torch.randn(3, 40, 18)
    x_lens = torch.randint(30, 41, size=(3,), dtype=torch.int32)
    x_lens[-1] = 40
    y = torch.randint(0, 100, size=(3, 10), dtype=torch.int32)
    y_lens = torch.randint(0, 11, size=(3,), dtype=torch.int32)
    y_lens[-1] = 10  # x_lens 和 y_lens必须有一个值顶到最大长度，不然算rnnt_loss时会报错

    encoder = Tdnn(
        input_size=18,
        output_size=32,
        num_layers=3,
        output_size_list=[32, 32, 32],
        kernel_size_list=[3, 5, 5],
        dilation_size_list=[1, 2, 4],
    )
    decoder = LSTMDecoder(
        vocab_size=100,
        num_layers=3,
        hidden_dim=32,
        embedding_dim=32,
        blank_id=0,
        embedding_dropout=0.01,
        rnn_dropout=0.01
    )
    joiner = Joiner(32, 100)
    model = Transducer(encoder, decoder, joiner)
    loss = model(x, x_lens, y, y_lens)
    print(loss)

    res_list = model.decode(x, x_lens)
    print(res_list)
    print(len(res_list))
    print(len(res_list[0]))
    print(len(res_list[1]))
    print(len(res_list[2]))


if __name__ == '__main__':
    test_transducer()



import requests
import argparse
import gxl_ai_utils

# url和端口携程自己的
flask_url = 'http://127.0.0.1:5012/predict'


def predict_result(image_path):
    #啥方法都行
    image = open(image_path, 'rb').read()
    payload = {'gxl_img': image}
    #request发给server.
    r = requests.post(flask_url, files=payload).json()

    # 成功的话在返回.
    if r['success']:
        # 输出结果.
        for (i, result) in enumerate(r['predictions']):
            print('{}. {}: {:.4f}'.format(i + 1, result['label'],
                                          result['probability']))
    # 失败了就打印.
    else:
        print('Request failed')


if __name__ == '__main__':
    data_path = gxl_ai_utils.AiConstant.DATA_PATH
    parser = argparse.ArgumentParser(description='Classification demo')
    parser.add_argument('--file', default=data_path+'/flower_data/train_filelist/image_06998.jpg', type=str, help='test_gxl_ai_utils image file')

    args = parser.parse_args()
    predict_result(args.file)



import gzip
import os
import pickle
from abc import ABC
from pathlib import Path

import numpy as np
from PIL import Image
from torchvision import datasets, transforms
from torch.utils.data import DataLoader
import requests
import torch.utils.data
from torch.utils.data import Dataset, DataLoader
from gxl_ai_utils.AiConstant import DATA_PATH

from . import store_dataloader
from . import store_data_name
from .store_dataset import store_dataset


class DataStore:
    """数据仓库
    如果指定数据集的位置, 则代表已经自己下载数据集, 如果没有指定位置, 则代表还没有下载, 程序自动下载到默认目录,
    下载的位置不能自己指定
    """

    def __init__(self, data_name, data_path=None):
        self.data_name = data_name
        self.data_path = data_path
        self.PATH = Path(DATA_PATH) / data_name if data_path is None else Path(data_path)
        self.str_path = str(self.PATH)
        self.PATH.mkdir(parents=True, exist_ok=True)

    def get_dataset(self):
        if self.data_name == store_data_name.MNIST:

            pipeline = transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,))
            ])
            # 下载数据集

            train_set = datasets.MNIST(root=self.str_path, train=True,
                                       download=self.data_path is None, transform=pipeline)
            test_set = datasets.MNIST(root=self.str_path, train=False,
                                      download=self.data_path is None, transform=pipeline)
            return train_set, test_set
        elif self.data_name == store_data_name.MNIST_RANDOM:
            # 定义一个随机数据集
            return store_dataset.RandomDataset(50000, 1, 28), store_dataset.RandomDataset(10000, 1, 28)
        elif self.data_name == store_data_name.FLOWER_CLASSIFY:
            data_dir = self.PATH
            train_dir = data_dir / 'handled' / 'train'
            valid_dir = data_dir / 'handled' / 'valid'
            data_transforms = store_dataset.FlowerDataset.data_transforms
            image_datasets = {x: datasets.ImageFolder(os.path.join(data_dir, x), data_transforms[x]) for x in
                              ['train', 'valid']}
            return image_datasets['train'], image_datasets['valid']
        elif self.data_name == store_data_name.FLOWER_CLASSIFY_SELF_DATASET:
            train_set = store_dataset.FlowerDataset(
                resource_dir=self.str_path + '/flower_data/train_filelist',
                mapping_list_file=self.str_path + '/flower_data/train.txt',
                transform=store_dataset.FlowerDataset.data_transforms['train']
            )
            valid_set = store_dataset.FlowerDataset(
                resource_dir=self.str_path + '/flower_data/val_filelist',
                mapping_list_file=self.str_path + '/flower_data/val.txt',
                transform=store_dataset.FlowerDataset.data_transforms['valid']
            )
            return train_set, valid_set
        elif self.data_name == store_data_name.TEXT_CLASSIFY:
            return store_dataset.get_text_classify_dataset(self.str_path)
        else:
            print('找不到数据集')
            return None

    def get_dataloader(self, batch_size=64):
        X = self.get_dataset()
        if len(X) == 2:
            train_set, test_set = X
            train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
            test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=True)
            return train_loader, test_loader
        elif len(X) == 3:
            train_set, valid_set, test_set = X
            train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
            valid_loader = DataLoader(valid_set, batch_size=batch_size, shuffle=True)
            test_loader = DataLoader(test_set, batch_size=batch_size, shuffle=True)
            return train_loader, valid_loader, test_loader
        elif len(X) == 1:
            train_set = X
            train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
            return train_loader
        else:
            return None



import copy
import datetime
import sys
from contextlib import nullcontext


import torch
from gxl_ai_utils.training.trainer.utils_for_excutor import do_training_join, do_batch_forward, do_batch_backward, \
    do_update_parameter_and_lr, log_per_step, save_model
from gxl_ai_utils.training.trainer.utils_for_excutor import GxlStepTimer
from gxl_ai_utils.utils import utils_file
class Executor:
    def __init__(self, global_step:int=0,
                 device: torch.device = torch.device('cpu')):
        """"""
        self.step = global_step + 1
        self.train_step_timer = None
        self.cv_step_timer = None
        self.device = device

    def train(self,
              model,
              optimizer,
              scheduler,
              scaler,
              configs,
              writer,
              group_join,
              train_data_loader,
              cv_data_loader):
        """
        训一个epoch
        :return:
        """
        if self.train_step_timer is None:
            self.train_step_timer = GxlStepTimer(self.step)
        model.train()
        info_dict = copy.deepcopy(configs)
        utils_file.logging_info('使用梯度累计， 新的batch_size是之前的{}倍'.format(info_dict['accum_grad']))
        if isinstance(model, torch.nn.parallel.DistributedDataParallel):
            model_context = model.join
        else:
            model_context = nullcontext
        info_dict['tag'] = 'TRAIN'
        with model_context():
            for batch_idx, batch in enumerate(train_data_loader):
                info_dict['step'] = self.step
                info_dict['batch_idx'] = batch_idx
                if do_training_join(group_join, info_dict):
                    break
                if batch['target_lengths'].size(0) == 0:
                    continue
                if info_dict['train_engine'] in ['torch_ddp','torch_fsdp'] and (batch_idx+1)%info_dict['accum_grad']!=0:
                    no_sync_context = model.no_sync # 不到梯度累计的时候，就不进行梯度同步
                else:
                    no_sync_context = nullcontext
                with no_sync_context():
                    info_dict = do_batch_forward(model, batch, scaler, info_dict, self.device)
                    info_dict = do_batch_backward(model, scaler, info_dict)
                # loss.backward()反向传导后才具有梯度，但是并不会更改参数，需要自己手动根据梯度进行参数更改，也就是该函数所做的事情
                info_dict = do_update_parameter_and_lr(model, optimizer, scheduler, scaler, info_dict)
                log_per_step(writer, info_dict, timer=self.train_step_timer)
                save_interval = info_dict.get('save_interval', sys.maxsize)
                if (self.step +
                    1) % save_interval == 0 and self.step != 0 and (
                        batch_idx + 1) % info_dict["accum_grad"] == 0:
                    import torch.distributed as dist
                    # Ensure all ranks start CV at the same time in step mode
                    dist.barrier()
                    loss_dict = self.cv(model, cv_data_loader, configs)
                    model.train()
                    info_dict.update({
                        "tag":
                            "step_{}".format(self.step),
                        "loss_dict":
                            loss_dict,
                        "save_time":
                            datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S'),
                        "lrs":
                            [group['lr'] for group in optimizer.param_groups]
                    })
                    save_model(model, info_dict)
                    # write final cv: tensorboard
                    log_per_step(writer, info_dict)
                    # Ensure all ranks start Train at the same time in step mode
                    dist.barrier()
                self.step += 1 if (batch_idx +
                                   1) % info_dict["accum_grad"] == 0 else 0






    def cv(self, model, cv_data_loader, configs):
        """"""
        if self.cv_step_timer is None:
            self.cv_step_timer = GxlStepTimer(0.0)
        else:
            self.cv_step_timer.last_iteration = 0.0
        model.eval()
        info_dict = copy.deepcopy(configs)
        num_seen_utts, loss_dict, total_acc = 1, {}, []  # avoid division by 0
        with torch.no_grad():
            for batch_idx, batch_dict in enumerate(cv_data_loader):
                info_dict["tag"] = "CV"
                info_dict["step"] = self.step
                info_dict["batch_idx"] = batch_idx
                info_dict["cv_step"] = batch_idx

                num_utts = batch_dict["target_lengths"].size(0)
                if num_utts == 0:
                    continue

                info_dict = do_batch_forward(model, batch_dict,  info_dict,None,
                                          self.device)
                _dict = info_dict["loss_dict"]

                num_seen_utts += num_utts
                total_acc.append(_dict['th_accuracy'].item(
                ) if _dict.get('th_accuracy', None) is not None else 0.0)
                for loss_name, loss_value in _dict.items():
                    if loss_value is not None and "loss" in loss_name \
                            and torch.isfinite(loss_value):
                        loss_value = loss_value.item()
                        loss_dict[loss_name] = loss_dict.get(loss_name, 0) + \
                                               loss_value * num_utts
                # write cv: log
                log_per_step(writer=None,
                             info_dict=info_dict,
                             timer=self.cv_step_timer)
        for loss_name, loss_value in loss_dict.items():
            loss_dict[loss_name] = loss_dict[loss_name] / num_seen_utts
        loss_dict["acc"] = sum(total_acc) / len(total_acc)
        return loss_dict





from tqdm import tqdm

from gxl_ai_utils.audio_dataset.common_data_handler import *

import pickle
import torch
import torch.nn.utils.rnn as rnn
import os

from gxl_ai_utils.config.gxl_config import GxlNode

from sortedcontainers import SortedSet
from torch.utils.data import DataLoader


class LabelTokenizer:
    UNK_TOKEN = "<UNK>"
    UNK_ID = -1

    def __init__(self, vocab_file_path='./output/label_vocab_dict.pkl'):
        """"""
        self.vocab_file_path = vocab_file_path
        self._id_to_token = {}
        self._token_to_id = {}
        os.makedirs(os.path.dirname(vocab_file_path), exist_ok=True)

    def build_vocab(self, label_list):
        sortedSet = SortedSet(label_list)
        self._token_to_id = {token: index for index, token in enumerate(sortedSet)}
        self._token_to_id[self.UNK_TOKEN] = self.UNK_ID
        self._id_to_token = {token_id: token for token, token_id in self._token_to_id.items()}
        with open(self.vocab_file_path, "wb") as f:
            pickle.dump(list(sortedSet), f)

    def load_vocab(self, ):
        """加载已保存的字典"""
        with open(self.vocab_file_path, "rb") as f:
            sortedSet = SortedSet(pickle.load(f))
        self._token_to_id = {token: index for index, token in enumerate(sortedSet)}
        self._token_to_id[self.UNK_TOKEN] = self.UNK_ID
        self._id_to_token = {token_id: token for token, token_id in self._token_to_id.items()}

    def __len__(self):
        return len(self._token_to_id)

    def encode(self, x: str | list) -> list[int]:
        if isinstance(x, str):
            return [self._token_to_id.get(x, self.UNK_ID)]
        elif isinstance(x, list):
            return [self._token_to_id.get(token, self.UNK_ID) for token in x]

    def decode(self, x: int | list | torch.Tensor) -> list[str]:
        """解码，输入int, int_list, int_tensor，只支持一维ids
        return -> token list"""
        if isinstance(x, int):
            return [self._id_to_token.get(x, self.UNK_TOKEN)]
        if isinstance(x, list):
            assert isinstance(x[0], int)
            return [self._id_to_token.get(token_id, self.UNK_TOKEN) for token_id in x]
        elif isinstance(x, torch.Tensor):
            assert x.dim() == 1
            return [self._id_to_token.get(token_id, self.UNK_TOKEN) for token_id in x]

    def show_token_id(self):
        for token, id in self._token_to_id.items():
            print(token + ' --- ' + str(id))


class ChineseVocabTokenizer:
    SOD_TOKEN = "<SOD>"
    EOD_TOKEN = "<EOD>"
    UNK_TOKEN = "<UNK>"
    PAD_TOKEN = "<PAD>"
    BLANK_TOKEN = "_BLANK_"
    SOD_ID = 0
    EOD_ID = 1
    UNK_ID = 2
    PAD_ID = 3
    BLANK_ID = 4

    def __init__(self, tokenizer=lambda x: [y for y in x], vocab_file_path='./output/vocab_set.pkl'):
        self.vocab_set = SortedSet()
        self._token_to_id = {}
        self._id_to_token = {}
        self.tokenizer = tokenizer
        self.vocab_file_path = vocab_file_path

    def load_vocab(self, ):
        """加载已保存的字典"""
        with open(self.vocab_file_path, "rb") as f:
            self.vocab_set = SortedSet(pickle.load(f))
        self._from_set_build_dict()

    def build_vocab(self, label_list):
        """构建字典"""
        for label in label_list:
            self.vocab_set.update(self.tokenizer(label))
        self._from_set_build_dict()
        with open(self.vocab_file_path, "wb") as f:
            pickle.dump(list(self.vocab_set), f)

    def _from_set_build_dict(self):
        self._token_to_id = {token: id + 5 for id, token in enumerate(self.vocab_set)}
        self._token_to_id[self.SOD_TOKEN] = self.SOD_ID
        self._token_to_id[self.EOD_TOKEN] = self.EOD_ID
        self._token_to_id[self.UNK_TOKEN] = self.UNK_ID
        self._token_to_id[self.PAD_TOKEN] = self.PAD_ID
        self._token_to_id[self.BLANK_TOKEN] = self.BLANK_ID
        self._id_to_token = {token_id: token for token, token_id in self._token_to_id.items()}

    def __len__(self):
        return len(self._token_to_id)

    def encode(self, x):
        """x: str, list, 只能是一维的token_list
        return -> id_list"""
        if isinstance(x, str) and self._token_to_id.get(x, None):
            return [self._token_to_id[x]]  # 如果为单个token
        if isinstance(x, str):
            x = self.tokenizer(x)
        assert isinstance(x, list)
        if len(x) == 0:
            return []
        assert isinstance(x[0], str)
        return [self._token_to_id.get(token, self.UNK_ID) for token in x]

    def decode(self, x: int | list | torch.Tensor):
        """解码，输入int, int_list, int_tensor，只支持一维ids
        return -> token list"""
        if isinstance(x, int):
            return [self._id_to_token.get(x, self.UNK_TOKEN)]
        if isinstance(x, list):
            assert isinstance(x[0], int)
            return [self._id_to_token.get(token_id, self.UNK_TOKEN) for token_id in x]
        elif isinstance(x, torch.Tensor):
            assert x.dim() == 1
            return [self._id_to_token.get(token_id, self.UNK_TOKEN) for token_id in x]

    def show_token_id(self):
        for token, id in self._token_to_id.items():
            print(token + ' --- ' + str(id))


def build_tokenizer_by_jsonl(jsonl_file, config: GxlNode):
    _, label_list = get_wavfilepath_and_label_from_jsonl(jsonl_file, config.dataset.wav_key, config.dataset.label_key)
    tokenizer = ChineseVocabTokenizer()
    tokenizer.build_vocab(label_list)
    tokenizer.show_token_id()


def build_tokenizer_by_json(json_file, config: GxlNode):
    _, label_list = get_wavfilepath_and_label_from_json(json_file, config.dataset.wav_key, config.dataset.label_key)
    tokenizer = ChineseVocabTokenizer()
    tokenizer.build_vocab(label_list)
    tokenizer.show_token_id()


def load_tokenizer():
    tokenizer = ChineseVocabTokenizer()
    tokenizer.load_vocab()
    return tokenizer


class RecognitionDataset(GxlAudioDataset):
    def __init__(self, wav_path_list, label_list, config: GxlNode):
        super(RecognitionDataset, self).__init__(wav_path_list, label_list, config)
        self.tokenizer = load_tokenizer()
        if config.dataset.global_cmvn_file:
            cmvn_dict = utils_file.load_dict_from_json(config.dataset.global_cmvn_file)
            self.global_means = torch.tensor(cmvn_dict['global_means'])
            self.global_vars = torch.tensor(cmvn_dict['global_vars'])

    def __getitem__(self, index):
        mel_num = self.config.dataset.mel_num
        min_sec = self.config.dataset.min_sec
        max_sec = self.config.dataset.max_sec
        label = self.tokenizer.encode(self.label_list[index])
        wav = compute_fbank(self.wav_path_list[index], mel_num=mel_num, min_sec=min_sec, max_sec=max_sec)
        if wav is None:
            return None, None
        if self.global_vars is not None:
            wav = apply_cmvn(self.global_means, self.global_vars, wav)
        return wav, torch.tensor(label, dtype=torch.int32)


def compute_cmvn(fbank_feature: torch.Tensor):
    """
    :param fbank_feature: tensor:(frame_num, mel_len)
    """
    mean = torch.mean(fbank_feature, dim=0)
    var = torch.var(fbank_feature, dim=0)
    return mean, var


def calculate_fbank_cmvn(json_file: str, config: GxlNode):
    """"""
    if config.dataset.global_cmvn_file is None:
        return
    os.makedirs(os.path.dirname(config.dataset.global_cmvn_file), exist_ok=True)
    big_dic = utils_file.load_dict_from_json(json_file)
    means = []
    vars = []
    values = list(big_dic.values())
    for dic in tqdm(values, total=len(values)):
        wav_path = dic[config.dataset.wav_key]
        fbank_feature = compute_fbank(wav_path, config.dataset.mel_num, config.dataset.min_sec, config.dataset.max_sec)
        if fbank_feature is None:
            continue
        mean, var = compute_cmvn(fbank_feature)
        means.append(mean)
        vars.append(var)
    means_tensor = torch.stack(means)
    vars_tensor = torch.stack(vars)
    global_means = torch.mean(means_tensor, dim=0)
    global_vars = torch.mean(vars_tensor, dim=0)
    res_dict = {"global_means": global_means.tolist(), "global_vars": global_vars.tolist()}
    utils_file.write_dict_to_json(res_dict, config.dataset.global_cmvn_file, )


def apply_cmvn(mean: torch.Tensor, var: torch.Tensor, fbank_feature: torch.Tensor):
    return (fbank_feature - mean) / var


def collate_fn(batch):
    # batch 是一个列表，每个元素是一个元组 (wav, label)
    batch = [(wav, label) for (wav, label) in batch if wav is not None and label is not None]
    data_list, target_list = zip(*batch)
    data_lens = torch.tensor([len(item) for item in data_list], dtype=torch.int32)
    wav_input = rnn.pad_sequence(data_list, batch_first=True)
    target_lens = torch.tensor([len(item) for item in target_list], dtype=torch.int32)
    target_input = rnn.pad_sequence(target_list, batch_first=True)
    return wav_input, data_lens, target_input, target_lens


def get_iter_by_json(json_file_path, config: GxlNode):
    dataset = get_dataset_by_json(json_file_path, config)
    dataloader = DataLoader(dataset, batch_size=config.dataset.batch_size, drop_last=config.dataset.drop_last,
                            shuffle=config.dataset.shuffle,
                            collate_fn=collate_fn, num_workers=config.dataset.num_workers,
                            pin_memory=config.dataset.pin_memory)
    return dataloader


def get_iter_by_jsonl(jsonl_file_path, config: GxlNode):
    dataset = get_dataset_by_jsonl(jsonl_file_path, config)
    dataloader = DataLoader(dataset, batch_size=config.dataset.batch_size, drop_last=config.dataset.drop_last,
                            shuffle=config.dataset.shuffle,
                            collate_fn=collate_fn, num_workers=config.dataset.num_workers,
                            pin_memory=config.dataset.pin_memory)
    return dataloader


def get_dataset_by_json(json_file_path, config: GxlNode):
    wav_path_list, label_list = get_wavfilepath_and_label_from_json(json_file_path, config.dataset.wav_key,
                                                                    config.dataset.label_key)
    dataset = RecognitionDataset(wav_path_list, label_list, config)
    return dataset


def get_dataset_by_jsonl(json_file_path, config: GxlNode):
    wav_path_list, label_list = get_wavfilepath_and_label_from_jsonl(json_file_path, config.dataset.wav_key,
                                                                     config.dataset.label_key)
    dataset = RecognitionDataset(wav_path_list, label_list, config)
    return dataset



import os
import time
import traceback
from enum import Enum

try:
    import requests
    from lxml import etree
    from selenium import webdriver
    from selenium.webdriver.chrome.options import Options
    from selenium.webdriver.common.by import By
except Exception as e:
    print(e)

COMMON_HEADER = {
    "Accept": "text.txt/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "zh-CN,zh;q=0.9,en;q=0.8",
    "Cache-Control": "max-age=0",
    "Connection": "keep-alive",
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "same-origin",
    "Sec-Fetch-User": "?1",
    "Upgrade-Insecure-Requests": "1",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": "Windows",
}


class REQUEST_METHOD(Enum):
    GET = 1
    POST = 2


def send_request(src, headers: dict = None, encoding='utf-8', method: REQUEST_METHOD = REQUEST_METHOD.GET):
    """
    发送请求, 如果顺利得到response并没有爆出任何错误,同时状态码为200, 则正常返回, 否则给出失败提示,
    需要调用者自行决定是否进行成功提示, 该函数不提供
    """
    if headers is None:
        headers = COMMON_HEADER
    response = None
    try:
        if method == REQUEST_METHOD.GET:
            response = requests.get(src, headers=headers)
        elif method == REQUEST_METHOD.POST:
            response = requests.post(src, headers=headers)
        if not hasattr(response, 'status_code'):
            raise RuntimeError()
        if not (str(response.status_code).startswith('2')):
            raise RuntimeError()
    except Exception as e:
        if hasattr(response, 'status_code'):
            print(f"请求失败{e},code:{response.status_code}")
        else:
            print(f"请求失败{e}")
        traceback.print_exc()
        return None
    # response.encoding = encoding
    # # 尝试从响应头中获取编码信息
    # encoding = response.encoding
    # # 如果没有指定编码，尝试使用自动检测编码
    # if not encoding:
    #     response.encoding = response.apparent_encoding
    # response.encoding = encoding
    return response



class FileType(Enum):
    HTML = 1,
    XML = 2,


def text2special_file(text, file_type: FileType = FileType.HTML):
    if file_type == FileType.HTML:
        try:
            tree = etree.HTML(text)
        except Exception as e:
            print('text.txt->html失败')
            return None
        return tree


def write_to_html(data, file_name):
    if not os.path.exists(os.path.dirname(file_name)):
        os.makedirs(os.path.dirname(file_name), exist_ok=True)
    with open(file_name, 'w', encoding='utf-8') as f:
        f.write(data)


def handle_xpath(html_text: str, xpath: str):
    try:
        tree = etree.HTML(html_text)
        datas = tree.xpath(xpath)
        # datas = tree.cssselect(xpath)
        if len(datas) == 0:
            print("warning:datas为空, 没得到应有的数据:", datas)
        return datas
    except Exception as e:
        print('xpath处理失败')
        traceback.print_exc()
        return None


def download_file(url: str, file_path: str):
    response = send_request(url)
    with open(file_path, 'wb') as f:
        f.write(response.content)


def handle_css_select(html_text: str, css_selector: str):
    try:
        tree = etree.HTML(html_text)
        datas = tree.cssselect(css_selector)
        # datas = tree.cssselect(xpath)
        if len(datas) == 0:
            print("datas:", datas)
            raise RuntimeError()
        return datas
    except Exception as e:
        print('xpath处理失败')
        traceback.print_exc()
        return None


def hande_href(href: str, xpath: str, mode='xpath'):
    if mode == 'xpath':
        return handle_xpath(send_request(href).text, xpath)
    elif mode == 'css':
        return handle_css_select(send_request(href).text, xpath)
    # return handle_xpath(send_request(href).text.txt, xpath)


def hande_href_by_brower(href: str, xpath: str, is_debug_mode=False):
    text, _ = get_source_page_from_url(href, is_debug_mode)
    return handle_xpath(text, xpath)


def handle_href_for_two(href: str, xpath1: str, xpath2: str):
    """
    传入两个xpath，针对同一块的不同部分，要求两个xpath对应的数据数量相同
    :param href:
    :param xpath1:
    :param xpath2:
    :return:
    """
    res_data = hande_href(href, xpath1)
    res_text = hande_href(href, xpath2)
    res = []
    for data, text in zip(res_data, res_text):
        res.append((data, text))
    return res


def handle_xpath_sub(tree: etree.ElementTree, xpath):
    try:
        datas = tree.xpath(xpath)
        if len(datas) == 0 or len(datas[-1]) == 0:
            raise RuntimeError()
        return datas
    except Exception as e:
        print('xpath处理失败')
        return None


def do_a_click_to_a_driver(driver: webdriver.Chrome, button_location_xpath):
    """
    传入一个button的地址, 做到点击它的效果, 并返回页面源码
    """
    button = driver.find_element(By.XPATH, button_location_xpath)
    button.click()
    driver.implicitly_wait(1)
    time.sleep(1)
    # 获取网页源码
    html = driver.page_source
    return html, driver

def get_a_driver(is_debug_chrome=False):
    if is_debug_chrome:
        options = Options()
        options.add_experimental_option("debuggerAddress", "127.0.0.1:9222")
        driver = webdriver.Chrome(options=options)
    else:
        driver = webdriver.Chrome()
    return driver
def get_source_page_from_url(url: str, is_debug_chrome: bool = False,wait_time=2):
    #如果采用is_debug_chrome模式，则先在cmd中运行该命令：
    # chrome.exe --remote-debugging-port=9222 --user-gxl_data-dir="D:\temp"
    if is_debug_chrome:
        options = Options()
        options.add_experimental_option("debuggerAddress", "127.0.0.1:9222")
        driver = webdriver.Chrome(options=options)
    else:
        driver = webdriver.Chrome()
    driver.get(url)
    driver.implicitly_wait(wait_time)
    time.sleep(wait_time)
    html = driver.page_source
    return html, driver


def get_source_page_from_url_have_driver(url: str, driver: webdriver.Chrome):
    driver.get(url)
    driver.implicitly_wait(1)
    time.sleep(1)
    html = driver.page_source
    return html, driver


def load_from_html(param):
    with open('text.html', 'r', encoding='utf-8') as file:
        html_content = file.read()
    return html_content


# ddp_demo.py
import torch
import torch.nn as nn
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
import os
import argparse

import gxl_ai_utils
import runner
from torch import distributed, optim
from config import get_logger
import config
import tokenize_vocab_word
from chat_model_module_rnn import GxlChatRnnModel
from data_handle import get_loss, get_dataset

os.environ['OMP_NUM_THREADS'] = "1"

parser = argparse.ArgumentParser()
parser.add_argument('--gpu_id', type=str, default='0,1,2')  # 不设置默认值也不赋值的话会爆出找不到gpus的错误
parser.add_argument('--batchSize', type=int, default=612)
parser.add_argument('--epochs', type=int, default=15)
config_parser = parser.parse_args()

os.environ['CUDA_VISIBLE_DEVICES'] = config_parser.gpu_id
torch.distributed.init_process_group(backend="nccl")

local_rank = torch.distributed.get_rank()
torch.cuda.set_device(
    local_rank)  # 只能设置local_rank，设为int(config_parser.gpu_id.split(',')[localrank]),
# 由于环境变量的设置，程序会使用gpu_id里面的设别，但是计数仍然是0,1,2...

# --------------------------------------------------------- start your code
tokenizer = tokenize_vocab_word.get_tokenizer()
net = GxlChatRnnModel(len(tokenizer))
#/model_params_150_2023-08-26_15:16:12.pth
params_with_prefix = torch.load(config.MODELSAVEPATH + "model_params_150_2023-08-26_15:16:12.pth",
                                map_location=torch.device('cuda:' + str(local_rank)))
params_without_prefix = {k.replace("module.", ""): v for k, v in params_with_prefix.items()}
net.load_state_dict(params_without_prefix)
optimizer = torch.optim.Adam(net.parameters(), lr=1e-4)
lr_schedule = lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=5, T_mult=2)
loss_func = get_loss()
train_dataset = get_dataset()
# -------------------------------------------------------- end your code


sampler1 = DistributedSampler(train_dataset)  # 这个sampler会自动分配数据到各个gpu上
loader1 = DataLoader(train_dataset, batch_size=config_parser.batchSize, sampler=sampler1)
# loader = DataLoader(dataset, batch_size=config.batchSize, shuffle=True)

if torch.cuda.is_available():
    net.cuda()
model = torch.nn.parallel.DistributedDataParallel(net)

runner_o = runner.RunnerGxl(model, optimizer, loss_func, loader1, logger=get_logger(), valid_loader=None, multi=True,
                            local_rank=local_rank)
runner_o.run(config_parser.epochs)



from gxl_ai_utils.audio_dataset.common_data_handler import *

import pickle
import torch
import torch.nn.utils.rnn as rnn
import os

from gxl_ai_utils.config.gxl_config import GxlNode

from sortedcontainers import SortedSet
from torch.utils.data import Dataset, DataLoader


class LabelTokenizer:
    UNK_TOKEN = "<UNK>"
    UNK_ID = -1

    def __init__(self, vocab_file_path='./output/label_vocab_dict.pkl'):
        """"""
        self.vocab_file_path = vocab_file_path
        self._id_to_token = {}
        self._token_to_id = {}
        os.makedirs(os.path.dirname(vocab_file_path), exist_ok=True)

    def build_vocab(self, label_list):
        sortedSet = SortedSet(label_list)
        self._token_to_id = {token: index for index, token in enumerate(sortedSet)}
        self._token_to_id[self.UNK_TOKEN] = self.UNK_ID
        self._id_to_token = {token_id: token for token, token_id in self._token_to_id.items()}
        with open(self.vocab_file_path, "wb") as f:
            pickle.dump(list(sortedSet), f)
        print('vocab build done!')
        self.show_token_id()

    def load_vocab(self, ):
        """加载已保存的字典"""
        with open(self.vocab_file_path, "rb") as f:
            sortedSet = SortedSet(pickle.load(f))
        self._token_to_id = {token: index for index, token in enumerate(sortedSet)}
        self._token_to_id[self.UNK_TOKEN] = self.UNK_ID
        self._id_to_token = {token_id: token for token, token_id in self._token_to_id.items()}
        print('vocab load done!')
        self.show_token_id()

    def __len__(self):
        return len(self._token_to_id)

    def encode(self, x: str | list) -> list[int]:
        if isinstance(x, str):
            return [self._token_to_id.get(x, self.UNK_ID)]
        elif isinstance(x, list):
            return [self._token_to_id.get(token, self.UNK_ID) for token in x]

    def decode(self, x: int | list | torch.Tensor) -> list[str]:
        """解码，输入int, int_list, int_tensor，只支持一维ids
        return -> token list"""
        if isinstance(x, int):
            return [self._id_to_token.get(x, self.UNK_TOKEN)]
        if isinstance(x, list):
            assert isinstance(x[0], int)
            return [self._id_to_token.get(token_id, self.UNK_TOKEN) for token_id in x]
        elif isinstance(x, torch.Tensor):
            assert x.dim() == 1
            return [self._id_to_token.get(token_id, self.UNK_TOKEN) for token_id in x]

    def show_token_id(self):
        for token, id in self._token_to_id.items():
            print(token + ' --- ' + str(id))


def build_tokenizer(label_list):
    tokenizer = LabelTokenizer()
    tokenizer.build_vocab(label_list)
    return tokenizer


def load_tokenizer():
    tokenizer = LabelTokenizer()
    tokenizer.load_vocab()
    return tokenizer


class ClassifyDataset(GxlAudioDataset):
    def __init__(self, wav_path_list, label_list, config: GxlNode):
        super(ClassifyDataset, self).__init__(wav_path_list, label_list, config)
        self.tokenizer = build_tokenizer(label_list)

    def __getitem__(self, index):
        mel_num = self.config.dataset.mel_num
        min_sec = self.config.dataset.min_sec
        max_sec = self.config.dataset.max_sec
        label = self.tokenizer.encode(self.label_list[index])[0]
        wav = compute_fbank(self.wav_path_list[index], mel_num=mel_num, min_sec=min_sec, max_sec=max_sec)
        if wav is None:
            return None, None
        return wav, torch.tensor(label, dtype=torch.int32)


def collate_fn(batch):
    # batch 是一个列表，每个元素是一个元组 (wav, label)
    batch = [(wav, label) for (wav, label) in batch if wav is not None and label is not None]
    data_list, target_list = zip(*batch)
    data_lens = torch.tensor([len(item) for item in data_list])
    wav_input = rnn.pad_sequence(data_list, batch_first=True)
    target_list = torch.tensor(target_list, dtype=torch.long)
    return wav_input, data_lens, target_list


def get_iter_by_json(json_file_path, config: GxlNode):
    wav_path_list, label_list = get_wavfilepath_and_label_from_json(json_file_path)
    dataset = ClassifyDataset(wav_path_list, label_list, config)
    dataloader = DataLoader(dataset, batch_size=config.dataset.batch_size, drop_last=True, shuffle=True,
                            collate_fn=collate_fn, num_workers=2, pin_memory=True)
    return dataloader


def get_iter_by_jsonl(jsonl_file_path, config: GxlNode):
    wav_path_list, label_list = get_wavfilepath_and_label_from_jsonl(jsonl_file_path)
    dataset = ClassifyDataset(wav_path_list, label_list, config)
    dataloader = DataLoader(dataset, batch_size=config.dataset.batch_size, drop_last=True, shuffle=True,
                            collate_fn=collate_fn, num_workers=2, pin_memory=True)
    return dataloader


def get_dataset_by_json(json_file_path, config: GxlNode):
    wav_path_list, label_list = get_wavfilepath_and_label_from_json(json_file_path)
    dataset = ClassifyDataset(wav_path_list, label_list, config)
    return dataset


def get_dataset_by_jsonl(json_file_path, config: GxlNode):
    wav_path_list, label_list = get_wavfilepath_and_label_from_jsonl(json_file_path)
    dataset = ClassifyDataset(wav_path_list, label_list, config)
    return dataset



from setuptools import setup, find_packages

# 使用 find_packages 查找包
packages = find_packages(where='gxl_ai_utils')

# 打印所有找到的包，格式化为 'gxl_ai_utils/A', 'gxl_ai_utils/B', 'gxl_ai_utils/C' 的样式
formatted_packages = [f"gxl_ai_utils/{package}" for package in packages]
print("Found packages:", formatted_packages)

setup(
    name='gxl_ai_utils',
    version='1.5.6',
    author='Xuelong Geng',
    description='这个是耿雪龙的工具包模块, update time: 2025-10-13',
    author_email='3349495429@qq.com',
    packages=formatted_packages,  # 在安装时使用的包列表
    install_requires=[  # 安装依赖库
        'jsonlines',
        'colorama',
        'tqdm',
    ],
    package_dir={'': '.'},  # 设置根目录为.
)



from typing import Tuple, Optional

import torch
from torch import nn


class LSTMDecoder(nn.Module):
    def __init__(self,
                 hidden_dim: int,
                 embedding_dim: int,
                 vocab_size: int,
                 num_layers: int,
                 blank_id: int,
                 rnn_dropout: float,
                 embedding_dropout: float):
        """"""
        super(LSTMDecoder, self).__init__()
        self.embedding = nn.Embedding(
            num_embeddings=vocab_size,
            embedding_dim=embedding_dim,
            padding_idx=blank_id,
        )
        self.embedding_dropout = nn.Dropout(embedding_dropout)
        self.rnn = nn.LSTM(
            input_size=embedding_dim,
            hidden_size=hidden_dim,
            num_layers=num_layers,
            batch_first=True,
            dropout=rnn_dropout
        )
        self.blank_id = blank_id
        self.output_linear = nn.Linear(hidden_dim, hidden_dim)

    def forward(self,
                y: torch.Tensor,
                state: Optional[Tuple[torch.Tensor, torch]] = None,
                ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        embedding_out = self.embedding(y)
        embedding_out = self.embedding_dropout(embedding_out)
        lstm_out, (c, h) = self.rnn(embedding_out, state)
        lstm_out = self.output_linear(lstm_out)
        return lstm_out, (c, h)


def test_lstm_decoder():
    """"""
    model = LSTMDecoder(
        vocab_size=108,
        embedding_dim=128,
        hidden_dim=128,
        blank_id=0,
        rnn_dropout=0.01,
        embedding_dropout=0.01,
        num_layers=3
    )
    print(model)
    input_tensor = torch.LongTensor([[1, 5, 2, 103, 23, 35, 0, 0], [1, 5, 2, 103, 23, 35, 12, 100]])
    output_tensor, state = model(input_tensor)
    print(output_tensor.shape)
    print(state[0].shape)
    print(state[1].shape)
    output_tensor, state = model(input_tensor, state=state)
    print(output_tensor.shape)
    print(state[0].shape)
    print(state[1].shape)
